# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging

from .dim import DATE_KEYS
from .dim import DATETIME_KEYS
from .dim import LEVEL_KEYS
from .dim import MONTH_KEYS
from .dim import STEP_KEYS
from .dim import TIME_KEYS

LOG = logging.getLogger(__name__)


class Coord:
    def __init__(self, name, vals, dims=None, ds=None, component=None):
        self.name = name
        self.vals = vals
        self.dims = dims
        if not self.dims:
            self.dims = (self.name,)
        self.component = component

        if self.component:
            assert isinstance(self.component, tuple)
            assert len(self.component) == 2
            if len(self.vals) != len(self.component[1]):
                raise ValueError(
                    (
                        f"Invalid remapping for {name=}. Number of coordinate values"
                        " does not match number of component values,"
                        f" {len(self.vals)} != {len(self.component[1])}"
                    )
                )

    @staticmethod
    def make(name, *args, **kwargs):
        if name in DATETIME_KEYS:
            return DateTimeCoord(name, *args, **kwargs)
        elif name in DATE_KEYS:
            return DateCoord(name, *args, **kwargs)
        elif name in TIME_KEYS:
            return TimeCoord(name, *args, **kwargs)
        elif name in MONTH_KEYS:
            return MonthCoord(name, *args, **kwargs)
        elif name in STEP_KEYS:
            return StepCoord(name, *args, **kwargs)
        elif name in LEVEL_KEYS:
            return LevelCoord(name, *args, **kwargs)
        return Coord(name, *args, **kwargs)

    def to_xr_var(self, profile):
        import xarray

        # LOG.debug(f"{self.name=}" + str(self.convert(profile)))

        name = self.name
        return xarray.Variable(
            self.dims,
            self.convert(profile),
            self.attrs(name, profile),
        )

    def convert(self, profile):
        return self.vals

    def encoding(self, profile):
        return {}

    def attrs(self, name, profile):
        attrs = profile.attrs.coord_attrs.get(name, {})
        if profile.add_earthkit_attrs and self.component:
            attrs["_earthkit"] = {"keys": self.component[0], "values": self.component[1]}
        return attrs

    @staticmethod
    def _to_datetime_list(vals):
        import numpy as np

        from earthkit.data.utils.dates import to_datetime_list

        # datetime64 arrays are already in the required format
        if isinstance(vals, np.ndarray):
            if not np.issubdtype(vals.dtype, np.datetime64):
                return to_datetime_list(vals.tolist())
        else:
            return to_datetime_list(vals)

        return vals


class DateTimeCoord(Coord):
    def convert(self, profile):
        return Coord._to_datetime_list(self.vals)

    def attrs(self, name, profile):
        attrs = profile.attrs.coord_attrs.get(name, {})
        if self.component:
            attrs["_earthkit"] = {"keys": self.component[0]}
        return attrs


class DateCoord(Coord):
    def convert(self, profile):
        if profile.decode_times:
            return Coord._to_datetime_list(self.vals)
        return super().convert(profile)


class TimeCoord(Coord):
    def convert(self, profile):
        if profile.decode_timedelta:
            from earthkit.data.utils.dates import to_time
            from earthkit.data.utils.dates import to_timedelta

            return [to_timedelta(to_time(x)) for x in self.vals]
        return super().convert(profile)

    def encoding(self, profile):
        if profile.decode_timedelta:
            return ({"dtype": "timedelta64[s]"},)
        return {}


class StepCoord(Coord):
    resolution = None
    RESOLUTION_UNITS = {
        datetime.timedelta(hours=1): "hours",
        datetime.timedelta(minutes=1): "minutes",
        datetime.timedelta(seconds=1): "seconds",
    }

    def convert(self, profile):
        from earthkit.data.utils.dates import to_timedelta

        vals = [to_timedelta(x) for x in self.vals]

        if profile.decode_timedelta:
            return vals
        else:
            from earthkit.data.utils.dates import timedeltas_to_int

            vals, self.resolution = timedeltas_to_int(vals)

        return vals

    def encoding(self, profile):
        if profile.decode_timedelta:
            return ({"dtype": "timedelta64[s]"},)
        return {}

    def attrs(self, name, profile):
        attrs = super().attrs(name, profile)
        if self.resolution:
            if self.resolution in self.RESOLUTION_UNITS:
                attrs["units"] = self.RESOLUTION_UNITS[self.resolution]
            else:
                raise ValueError(f"Unsupported step resolution {self.resolution}")
        return attrs


class MonthCoord(Coord):
    def attrs(self, name, profile):
        attrs = super().attrs(name, profile)
        attrs["units"] = "months"
        return attrs


class LevelCoord(Coord):
    def __init__(self, name, vals, dims=None, ds=None, **kwargs):
        self.level_type = {}
        if ds is not None:
            for k in ["levtype", "typeOfLevel"]:
                v = ds[0].metadata(k, default=None)
                if v is not None:
                    self.level_type[k] = v

        super().__init__(name, vals, dims, **kwargs)

    def attrs(self, name, profile):
        attrs = profile.attrs
        conf = attrs.coord_attrs.get(name, {})
        if conf:
            key = conf["key"]
            if key in self.level_type:
                level_type = self.level_type[key]
                return conf.get(level_type, {})
        return {}
        # raise ValueError(f"Cannot determine level type for coordinate {name}")



import earthkit.data

# Access one of the config options
cache_path = earthkit.data.config.get("user-cache-directory")
print(cache_path)

# If this is the last line of a Notebook cell, this
# will display a table with all the current configuration
earthkit.data.config



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging

LOG = logging.getLogger(__name__)


def _match_magic(magic, deeper_check):
    if magic is not None and len(magic) > 0:
        type_id = b"GRIB"
        if not deeper_check:
            return len(magic) >= 4 and magic[:4] == type_id
        else:
            return type_id in magic
    return False


def _is_default(magic, content_type):
    return (magic is None or len(magic) == 0) and (content_type is None or len(content_type) == 0)


def is_grib_file(path):
    from earthkit.data.core.settings import SETTINGS

    n_bytes = SETTINGS.get("reader-type-check-bytes")
    magic = None
    with open(path, "rb") as f:
        magic = f.read(n_bytes)

    if _match_magic(magic, False):
        return True
    return _match_magic(magic, True)


def reader(source, path, *, magic=None, deeper_check=False, **kwargs):
    if _match_magic(magic, deeper_check):
        from .file import GRIBReader

        parts = source.parts if hasattr(source, "parts") else None
        return GRIBReader(source, path, parts=parts)


def memory_reader(source, buffer, *, magic=None, deeper_check=False, **kwargs):
    if _match_magic(magic, deeper_check):
        from .memory import GribFieldListInMemory
        from .memory import GribMessageMemoryReader

        return GribFieldListInMemory(source, GribMessageMemoryReader(buffer, **kwargs), **kwargs)


def stream_reader(
    source,
    stream,
    magic=None,
    *,
    deeper_check=False,
    content_type=None,
    memory=False,
    **kwargs,
):
    if _is_default(magic, content_type) or _match_magic(magic, deeper_check):
        from .memory import GribFieldListInMemory
        from .memory import GribStreamReader

        r = GribStreamReader(stream, **kwargs)
        if memory:
            r = GribFieldListInMemory(source, r, **kwargs)
        return r



import pytest

SKIP = {
    "short": ["download", "ftp", "long_test", "ecfs"],
    "long": ["ftp", "ecfs"],
    "release": [],
    "documentation": None,
}

# Import all fixtures from list of plugins
pytest_plugins = [
    "list_of_dicts.lod_fixtures",
    "patterns.patterns_fixtures",
]


def pytest_configure(config):
    pytest.CDS_TIMEOUT = config.getoption("--cds-timeout")


def pytest_addoption(parser):
    help_str = "NAME: short, long, release, documentation. Runs a subset of tests.\n"
    for k, v in SKIP.items():
        if v:
            help_str += f"'{k}': skip tests marked as {','.join(v)}.\n"
        else:
            help_str += f"'{k}': do not skip tests.\n"

    parser.addoption(
        "-E",
        action="store",
        metavar="NAME",
        default="short",
        help=help_str,
    )

    parser.addoption(
        "--cds-timeout",
        action="store",
        metavar="TIMEOUT",
        default=30,
        help="Timeout in seconds for cds tests. 0 means no timeout. Default is 30.",
    )


def pytest_runtest_setup(item):
    # print(f"config {item.config.option}")
    flag = item.config.getoption("-E")
    marks_to_skip = SKIP[flag]

    marks_in_items = list([m.name for m in item.iter_markers()])

    if marks_to_skip is None:
        if flag not in marks_in_items:
            pytest.skip(f"test is skipped because custom pytest option : -E {flag}")
        return

    for m in marks_in_items:
        if m in marks_to_skip:
            pytest.skip(f"test is skipped because custom pytest option: -E {flag}")

    marked_no_cache_init = "no_cache_init" in marks_in_items
    if marked_no_cache_init and not item.config.getoption("--forked"):
        pytest.skip("test is skipped because marked as no_cache_init but --forked is not used")

    need_cache = "cache" in marks_in_items

    # settings
    from earthkit.data import config

    # ensure settings are not saved automatically
    config.autosave = False

    # ensure all the tests use the default settings
    if marked_no_cache_init:
        # do not broadcast setting changes, otherwise
        # the cache would be initialised
        config._notify_enabled = False
        config.reset()
        config._notify_enabled = True
    elif need_cache:
        config.reset()
        config.set("cache-policy", "user")
    else:
        config.reset()



#!/usr/bin/env python3

# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import pytest

from earthkit.data.core.order import normalize_order_by

to_kwargs = normalize_order_by


def test_order_1():
    ref = {"date": "ascending", "time": "ascending", "level": "ascending"}

    assert to_kwargs(ref) == ref

    assert to_kwargs("date", "time", "level") == ref
    assert to_kwargs(["date", "time", "level"]) == ref
    assert to_kwargs(("date", "time", "level")) == ref


def test_order_2():
    ref = {"date": "descending", "time": "ascending", "level": "ascending"}

    assert to_kwargs(ref) == ref
    assert to_kwargs(dict(date="descending"), "time", "level") == ref
    assert to_kwargs(dict(date="descending"), ["time", "level"]) == ref


def test_order_3():
    ref = {"date": "ascending", "time": "ascending", "level": "descending"}

    assert to_kwargs(ref) == ref
    assert to_kwargs("time", "date", dict(level="descending")) == ref


def test_order_4():
    ref = {"date": None, "time": "ascending", "level": None}
    assert to_kwargs(dict(date=None), "time", dict(level=None)) == ref


def test_order_must_fail():
    with pytest.raises(ValueError):
        assert to_kwargs(5, 4)
    with pytest.raises(ValueError):
        assert to_kwargs([4, 5])
    with pytest.raises(ValueError):
        assert to_kwargs(a="ascending", b="blah")


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import numpy as np

LOG = logging.getLogger(__name__)


GEOGRAPHIC_COORDS = {
    "x": ["x", "X", "xc", "projection_x_coordinate", "lon", "longitude"],
    "y": ["y", "Y", "yc", "projection_y_coordinate", "lat", "latitude"],
}


class DataSet:
    """Class that wraps a xarray dataset to provide caching"""

    def __init__(self, ds):
        self._ds = ds
        self._bbox = {}
        self._cache = {}

    @property
    def data_vars(self):
        return self._ds.data_vars

    def __getitem__(self, name):
        return self._ds[name]

    def __getattr__(self, name):
        return getattr(self._ds, name)

    def bbox(self, variable):
        data_array = self[variable]

        keys, coords = self._get_xy_coords(data_array)
        key = ("bbox", tuple(keys), tuple(coords))
        if key in self._cache:
            return self._cache[key]

        # lons = self._get_xy(data_array, "x", flatten=False)
        # lats = self._get_xy(data_array, "y", flatten=False)

        lats, lons = self._get_latlon(data_array, flatten=False)

        north = np.amax(lats)
        west = np.amin(lons)
        south = np.amin(lats)
        east = np.amax(lons)

        self._cache[key] = (north, west, south, east)
        return self._cache[key]

    def _get_xy_coords(self, data_array):
        if (
            len(data_array.dims) >= 2
            and data_array.dims[-1] in GEOGRAPHIC_COORDS["x"]
            and data_array.dims[-2] in GEOGRAPHIC_COORDS["y"]
        ):
            return ("y", "x"), (data_array.dims[-2], data_array.dims[-1])

        keys = []
        coords = []
        axes = ("x", "y")
        for dim in data_array.dims:
            for ax in axes:
                candidates = GEOGRAPHIC_COORDS.get(ax, [])
                if dim in candidates:
                    keys.append(ax)
                    coords.append(dim)
                else:
                    ax = data_array.coords[dim].attrs.get("axis", "").lower()
                    if ax in axes:
                        keys.append(ax)
                        coords.append(dim)
            if len(keys) == 2:
                return tuple(keys), tuple(coords)

        for ax in axes:
            if ax not in keys:
                raise ValueError(f"No coordinate found with axis '{ax}'")

        return keys, coords

    def _get_xy(self, data_array, flatten=False, dtype=None):
        keys, coords = self._get_xy_coords(data_array)
        key = ("grid_points", tuple(keys), tuple(coords))

        if key in self._cache:
            points = self._cache[key]
        else:
            points = dict()
            v0, v1 = data_array.coords[coords[0]], data_array.coords[coords[1]]
            points[keys[1]], points[keys[0]] = np.meshgrid(v1, v0)
            self._cache[key] = points

        if flatten:
            points["x"] = points["x"].reshape(-1)
            points["y"] = points["y"].reshape(-1)

        if dtype is not None:
            return points["x"].astype(dtype), points["y"].astype(dtype)
        else:
            return points["x"], points["y"]

    def _get_latlon(self, data_array, flatten=False, dtype=None):
        keys, coords = self._get_xy_coords(data_array)

        points = dict()

        def _get_ll(keys):
            for key in keys:
                if key in self._ds:
                    return self._ds[key]

        lat_keys = ["latitude", "lat"]
        lon_keys = ["longitude", "lon"]
        latitude = _get_ll(lat_keys)
        longitude = _get_ll(lon_keys)

        if latitude is not None and longitude is not None:
            if latitude.dims == coords and longitude.dims == coords:
                latitude = latitude.data
                longitude = longitude.data
                points["y"] = latitude
                points["x"] = longitude

        if not points:
            key = ("grid_points", tuple(keys), tuple(coords))

            if key in self._cache:
                points = self._cache[key]
            else:
                v0, v1 = data_array.coords[coords[0]], data_array.coords[coords[1]]
                points[keys[1]], points[keys[0]] = np.meshgrid(v1, v0)
                self._cache[key] = points

        if flatten:
            points["x"] = points["x"].reshape(-1)
            points["y"] = points["y"].reshape(-1)

        if dtype is not None:
            return points["y"].astype(dtype), points["x"].astype(dtype)
        else:
            return points["y"], points["x"]



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pandas as pd
import xarray as xr

from earthkit.data import from_object
from earthkit.data import from_source
from earthkit.data.readers import Reader
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.utils import module_inputs_wrapper

from . import dummy_module
from .dummy_module import XR_TYPES

TEST_NP = np.arange(10)
TEST_NP2 = np.arange(10)

TEST_DF = pd.DataFrame({"index": TEST_NP, "data": TEST_NP2}).set_index("index")

TEST_DA = xr.DataArray(TEST_NP, name="test")
TEST_DA2 = xr.DataArray(TEST_NP2, name="test2")
TEST_DS = TEST_DA.to_dataset()
TEST_DS["test2"] = TEST_DA2

EK_XARRAY_WRAPPER = from_object(TEST_DS)
EK_NUMPY_WRAPPER = from_object(TEST_NP)

WRAPPED_XR_ONES_LIKE = module_inputs_wrapper.transform_function_inputs(
    xr.ones_like, kwarg_types={"other": XR_TYPES}
)

WRAPPED_XR_ONES_LIKE_TYPE_SETTING = module_inputs_wrapper.transform_function_inputs(
    dummy_module.xarray_ones_like,
)

WRAPPED_NP_MEAN = module_inputs_wrapper.transform_function_inputs(
    np.mean,
    kwarg_types={"a": np.ndarray},
    convert_types=(Reader),  # Only convert Earthkit.data.Reader (np.mean can handle xarray and pandas)
)

WRAPPED_NP_MEAN_TYPE_SETTING = module_inputs_wrapper.transform_function_inputs(
    dummy_module.numpy_mean,
)

WRAPPED_DUMMY_MODULE = module_inputs_wrapper.transform_module_inputs(
    dummy_module,
)


def test_transform_function_inputs_reader_to_xarray():
    # Check EK GribReader object
    EK_GRIB_READER = from_source("file", earthkit_test_data_file("test_single.grib"))
    ek_reader_result = WRAPPED_XR_ONES_LIKE(EK_GRIB_READER)
    # Will return a DataSet becuase that is first value in kwarg_types
    assert isinstance(ek_reader_result, xr.Dataset)
    assert ek_reader_result.equals(xr.ones_like(EK_GRIB_READER.to_xarray()))


def test_transform_function_inputs_reader_to_xarray_typesetting():
    # Check EK GribReader object
    EK_GRIB_READER = from_source("file", earthkit_test_data_file("test_single.grib"))
    ek_reader_result = WRAPPED_XR_ONES_LIKE_TYPE_SETTING(EK_GRIB_READER)
    # Will return a dataarray because that is first value in type-set Union
    assert isinstance(ek_reader_result, xr.DataArray)
    assert ek_reader_result.equals(xr.ones_like(EK_GRIB_READER.to_xarray()["2t"]))


def test_transform_module_inputs_reader_to_xarray():
    # Check EK GribReader object
    EK_GRIB_READER = from_source("file", earthkit_test_data_file("test_single.grib"))
    ek_reader_result = WRAPPED_DUMMY_MODULE.xarray_ones_like(EK_GRIB_READER)
    # Data array because type-setting of function has dataarray first
    assert isinstance(ek_reader_result, xr.DataArray)
    assert ek_reader_result.equals(xr.ones_like(EK_GRIB_READER.to_xarray())["2t"])


def test_transform_function_inputs_wrapper_to_xarray():
    # EK XarrayWrapper object
    ek_wrapper_result = WRAPPED_XR_ONES_LIKE(EK_XARRAY_WRAPPER)
    assert isinstance(ek_wrapper_result, xr.Dataset)
    assert ek_wrapper_result.equals(xr.ones_like(EK_XARRAY_WRAPPER.data))
    # EK NumpyWrapper object
    ek_wrapper_result = WRAPPED_XR_ONES_LIKE(EK_NUMPY_WRAPPER)
    assert isinstance(ek_wrapper_result, xr.DataArray)
    assert ek_wrapper_result.equals(xr.ones_like(TEST_DA))


def test_transform_module_inputs_wrapper_to_xarray():
    # EK XarrayWrapper object
    ek_wrapper_result = WRAPPED_DUMMY_MODULE.xarray_ones_like(EK_XARRAY_WRAPPER)
    # Will return a dataarray because that is first value in type-set Union
    assert isinstance(ek_wrapper_result, xr.DataArray)
    assert ek_wrapper_result.equals(xr.ones_like(EK_XARRAY_WRAPPER.data.test))
    # EK NumpyWrapper object
    ek_wrapper_result = WRAPPED_DUMMY_MODULE.xarray_ones_like(EK_NUMPY_WRAPPER)
    assert isinstance(ek_wrapper_result, xr.DataArray)
    assert ek_wrapper_result.equals(xr.ones_like(TEST_DA))


def test_transform_function_inputs_reader_to_numpy():
    # Test with Earthkit.data GribReader object
    EK_GRIB_READER = from_source("file", earthkit_test_data_file("test_single.grib"))
    assert WRAPPED_NP_MEAN(EK_GRIB_READER) == np.mean(EK_GRIB_READER.to_numpy())
    assert isinstance(WRAPPED_NP_MEAN(EK_GRIB_READER), np.float64)


def test_transform_function_inputs_reader_to_numpy_typesetting():
    # Test with Earthkit.data GribReader object
    EK_GRIB_READER = from_source("file", earthkit_test_data_file("test_single.grib"))
    result = WRAPPED_NP_MEAN_TYPE_SETTING(EK_GRIB_READER)
    assert result == np.mean(EK_GRIB_READER.to_numpy())
    assert isinstance(result, np.float64)


def test_transform_module_inputs_reader_to_numpy():
    # Test with Earthkit.data GribReader object
    EK_GRIB_READER = from_source("file", earthkit_test_data_file("test_single.grib"))
    result = WRAPPED_DUMMY_MODULE.numpy_mean(EK_GRIB_READER)
    assert result == np.mean(EK_GRIB_READER.to_numpy())
    assert isinstance(result, np.float64)


def test_transform_function_inputs_wrapper_to_numpy():
    # Test with Earthkit.data XarrayWrapper object
    ek_object_result = WRAPPED_NP_MEAN(EK_XARRAY_WRAPPER)
    assert ek_object_result == np.mean(TEST_DS)
    # assert isinstance(ek_object_result, type(EK_XARRAY_WRAPPER.data))

    # Test with Earthkit.data NumpyWrapper object
    ek_object_result = WRAPPED_NP_MEAN(EK_NUMPY_WRAPPER)
    assert ek_object_result == np.mean(TEST_NP)
    assert isinstance(ek_object_result, np.float64)


def test_transform_module_inputs_wrapper_to_numpy():
    # Test with Earthkit.data XarrayWrapper object
    ek_object_result = WRAPPED_DUMMY_MODULE.numpy_mean(EK_XARRAY_WRAPPER)
    assert ek_object_result == np.mean(TEST_DS)
    # TODO: Without definition of convert_types, WRAPPERS are converted to first arguement
    #       This could be addressed in future versions
    # assert isinstance(ek_object_result, type(EK_XARRAY_WRAPPER.data))

    # Test with Earthkit.data NumpyWrapper object
    ek_object_result = WRAPPED_DUMMY_MODULE.numpy_mean(EK_NUMPY_WRAPPER)
    assert ek_object_result == np.mean(TEST_NP)
    assert isinstance(ek_object_result, np.float64)


def test_transform_function_inputs_xarray_to_numpy():
    # Test with xarray.DataArray object
    assert WRAPPED_NP_MEAN(TEST_DA) == np.mean(TEST_NP)
    assert WRAPPED_NP_MEAN(TEST_DA) == np.mean(TEST_DA)
    assert isinstance(WRAPPED_NP_MEAN(TEST_DA), xr.DataArray)


def test_transform_function_inputs_pandas_to_numpy():
    # Test with pandas.DataFrame object, axis=0 is to preserve the pandas.DataFrame, this is pandas syntax
    result = WRAPPED_NP_MEAN(TEST_DF, axis=0)
    assert result.equals(np.mean(TEST_DF, axis=0))
    assert isinstance(result, pd.core.series.Series)

    # # Test without axis declartion, pandas>2 returns a scalar mean of the dataframe
    # # Omitted as breaking change in pandas makes tests unstable
    # result = WRAPPED_NP_MEAN(TEST_DF)
    # assert result == np.mean(TEST_DF)
    # assert isinstance(result, np.float64)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from . import EncodedData
from . import Encoder

LOG = logging.getLogger(__name__)


class GeoTIFFEncodedData(EncodedData):
    _GDAL_DRIVER = "GTiff"

    def __init__(self, ds):
        self.ds = ds

    def to_bytes(self, **kwargs):
        import rasterio.io

        m = rasterio.io.MemoryFile()
        self._to_raster(m, **kwargs)
        return m.getbuffer()

    def to_file(self, f, **kwargs):
        self._to_raster(f, **kwargs)

    def _to_raster(self, dst, **kwargs):
        options = {"driver": self._GDAL_DRIVER}
        options.update(kwargs)
        self.ds.rio.to_raster(dst, **options)

    def metadata(self, key):
        raise NotImplementedError


class GeoTIFFEncoder(Encoder):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def encode(self, data, **kwargs):
        if data is None:
            raise ValueError("No data to encode")

        from ..wrappers import get_wrapper

        data = get_wrapper(data)
        return data._encode(self, **kwargs)

    def _encode(self, data, **kwargs):
        ds = data.to_xarray()
        if ds.rio.crs is None:
            crs = data.projection().to_cartopy_crs()
            ds.rio.write_crs(crs, inplace=True)
        for var in ds.data_vars:
            if "_earthkit" in ds[var].attrs:
                del ds[var].attrs["_earthkit"]
        return GeoTIFFEncodedData(ds)

    def _encode_field(self, field, **kwargs):
        return self._encode(field, **kwargs)

    def _encode_fieldlist(self, fieldlist, **kwargs):
        return self._encode(fieldlist, **kwargs)

    def _encode_xarray(self, data, **kwargs):
        raise NotImplementedError


encoder = GeoTIFFEncoder



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from typing import Any as TypingAny
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

from earthkit.data.sources import Source
from earthkit.data.sources import from_source
from earthkit.data.sources.empty import EmptySource
from earthkit.data.sources.file import File
from earthkit.data.sources.multi import MultiSource
from earthkit.data.utils.patterns import HivePattern
from earthkit.data.utils.patterns import Pattern


class HiveFilePattern(Source):
    def __init__(self, pattern: str, params: Dict[str, TypingAny], **kwargs: TypingAny) -> None:
        self.scanner = HivePattern(pattern, params)

    def sel(
        self,
        *args: Tuple[Dict[str, TypingAny]],
        _hive_diag: Optional[TypingAny] = None,
        **kwargs: TypingAny,
    ) -> Union[EmptySource, MultiSource]:
        from earthkit.data.core.index import normalize_selection

        kwargs = normalize_selection(*args, **kwargs)

        rest = {k: v for k, v in kwargs.items() if k not in self.scanner.params}
        for k in rest:
            del kwargs[k]

        if rest:
            out = EmptySource()
            for f in self.scanner.scan(**kwargs):
                ds = from_source("file", f)
                out += ds.sel(**rest)
                if _hive_diag:
                    _hive_diag.file(1)
                    _hive_diag.sel(1)

            return out
        else:
            sources = [File(f) for f in self.scanner.scan(**kwargs)]

            if _hive_diag:
                _hive_diag.file(len(sources))

            src = MultiSource(sources)

            prev = None
            while src is not prev:
                prev = src
                src = src.mutate()
            return src


class FilePattern(MultiSource):
    def __init__(
        self,
        pattern: str,
        *args: Tuple[Dict[str, TypingAny]],
        filter: Optional[TypingAny] = None,
        merger: Optional[TypingAny] = None,
        hive_partitioning: bool = False,
        **kwargs: TypingAny,
    ) -> None:
        self.hive_partitioning = hive_partitioning

        if not self.hive_partitioning:
            files = Pattern(pattern).substitute(*args, **kwargs)
            if not isinstance(files, list):
                files = [files]

            sources = [File(file) for file in sorted(files)]
            super().__init__(sources, filter=filter, merger=merger)

        else:
            self.pattern = pattern
            params = {}
            for a in args:
                params.update(a)
            params.update(kwargs)
            self.params = params

    def mutate(self) -> Union["HiveFilePattern", "FilePattern"]:
        if self.hive_partitioning:
            return HiveFilePattern(self.pattern, self.params)
        else:
            return super().mutate()


source = FilePattern



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from xr_engine_fixtures import compare_dims  # noqa: E402


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,dims",
    [
        (
            {},
            {
                "number": [0, 1, 2],
            },
        ),
        (
            {
                "dim_roles": {"number": "perturbationNumber"},
                "dim_name_from_role_name": True,
            },
            {
                "number": [0, 1, 2],
            },
        ),
        (
            {
                "dim_roles": {"ens": "perturbationNumber"},
                "dim_name_from_role_name": True,
            },
            {
                "number": [0, 1, 2],
            },
        ),
        (
            {
                "dim_roles": {"number": "perturbationNumber"},
                "dim_name_from_role_name": False,
            },
            {
                "perturbationNumber": [0, 1, 2],
            },
        ),
    ],
)
def test_xr_number_dim(kwargs, dims):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/ens/ens_cf_pf.grib"))

    ds = ds_ek.to_xarray(**kwargs)
    compare_dims(ds, dims, order_ref_var="t")


@pytest.mark.cache
def test_xr_number_dim_missing():
    ds_ek = from_source(
        "url", earthkit_remote_test_data_file("test-data", "xr_engine", "date", "t2_td2_1_year.grib")
    )

    ds = ds_ek[10].to_xarray(time_dim_mode="valid_time", ensure_dims="number", fill_metadata={"number": 10})
    assert ds is not None

    dims = {"number": [10]}
    compare_dims(ds, dims, order_ref_var="2t")



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from array_fl_fixtures import load_array_fl_file  # noqa: E402

# Note: Almost all grib metadata tests are also run for numpyfs.
# See grib/test_grib_summary.py


def test_array_fl_dump():
    f, _ = load_array_fl_file("test6.grib")

    namespaces = (
        "default",
        "geography",
        "ls",
        "mars",
        "parameter",
        "time",
        "vertical",
    )

    # default
    r = f[0].dump(_as_raw=True)
    ref = [
        {
            "title": "ls",
            "data": {
                "edition": 1,
                "centre": "ecmf",
                "typeOfLevel": "isobaricInhPa",
                "level": 1000,
                "dataDate": 20180801,
                "stepRange": "0",
                "dataType": "an",
                "shortName": "t",
                "packingType": "grid_simple",
                "gridType": "regular_ll",
            },
            "tooltip": "Keys in the ecCodes ls namespace",
        },
        {
            "title": "geography",
            "data": {
                # "bitmapPresent": 0,
                # "Ni": 12,
                # "Nj": 7,
                "latitudeOfFirstGridPointInDegrees": 90.0,
                "longitudeOfFirstGridPointInDegrees": 0.0,
                "latitudeOfLastGridPointInDegrees": -90.0,
                "longitudeOfLastGridPointInDegrees": 330.0,
                "iScansNegatively": 0,
                "jScansPositively": 0,
                "jPointsAreConsecutive": 0,
                "jDirectionIncrementInDegrees": 30.0,
                "iDirectionIncrementInDegrees": 30.0,
                "gridType": "regular_ll",
            },
            "tooltip": "Keys in the ecCodes geography namespace",
        },
        {
            "title": "mars",
            "data": {
                "domain": "g",
                "levtype": "pl",
                "levelist": 1000,
                "date": 20180801,
                "time": 1200,
                "step": 0,
                "param": "t",
                "class": "od",
                "type": "an",
                "stream": "oper",
                "expver": "0001",
            },
            "tooltip": "Keys in the ecCodes mars namespace",
        },
        {
            "title": "parameter",
            "data": {
                "centre": "ecmf",
                "paramId": 130,
                "units": "K",
                "name": "Temperature",
                "shortName": "t",
            },
            "tooltip": "Keys in the ecCodes parameter namespace",
        },
        {
            "title": "time",
            "data": {
                "dataDate": 20180801,
                "dataTime": 1200,
                "stepUnits": 1,
                "stepType": "instant",
                "stepRange": "0",
                "startStep": 0,
                "endStep": 0,
                "validityDate": 20180801,
                "validityTime": 1200,
            },
            "tooltip": "Keys in the ecCodes time namespace",
        },
        {
            "title": "vertical",
            "data": {"typeOfLevel": "isobaricInhPa", "level": 1000},
            "tooltip": "Keys in the ecCodes vertical namespace",
        },
    ]

    assert len(r) == len(namespaces)
    assert isinstance(r, list)
    for d in r:
        ns = d["title"]
        assert ns in namespaces
        if ns == "geography":
            d["data"].pop("Ni", None)
            d["data"].pop("Nj", None)
            d["data"].pop("bitmapPresent", None)
        if ns not in ("default", "statistics"):
            assert d == [x for x in ref if x["title"] == ns][0], ns

    # a namespace
    r = f[0].dump(namespace="mars", _as_raw=True)
    ref = [
        {
            "title": "mars",
            "data": {
                "domain": "g",
                "levtype": "pl",
                "levelist": 1000,
                "date": 20180801,
                "time": 1200,
                "step": 0,
                "param": "t",
                "class": "od",
                "type": "an",
                "stream": "oper",
                "expver": "0001",
            },
            "tooltip": "Keys in the ecCodes mars namespace",
        }
    ]
    assert r == ref

    # namespace reformatted
    r = f[0].dump(namespace="mars", _as_raw=False)
    ref = {
        "mars": {
            "domain": "g",
            "levtype": "pl",
            "levelist": 1000,
            "date": 20180801,
            "time": 1200,
            "step": 0,
            "param": "t",
            "class": "od",
            "type": "an",
            "stream": "oper",
            "expver": "0001",
        }
    }
    assert r == ref


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import load_nc_or_xr_source


@pytest.mark.parametrize("mode", ["nc", "xr"])
@pytest.mark.parametrize(
    "params,expected_meta,metadata_keys",
    [
        (dict(variable="u", level=700), [["u", 700]], []),
        (
            dict(variable=["t", "u"], level=[700, 500]),
            [
                ["t", 700],
                ["t", 500],
                ["u", 700],
                ["u", 500],
            ],
            ["variable", "level"],
        ),
        (dict(variable="w"), [], []),
        (dict(INVALIDKEY="w"), [], []),
        (
            dict(
                variable=["t"],
                level=[500, 700],
                valid_datetime="2018-08-01T12:00:00",
            ),
            [
                ["t", 700, "2018-08-01T12:00:00"],
                ["t", 500, "2018-08-01T12:00:00"],
            ],
            ["variable", "level", "valid_datetime"],
        ),
        (
            dict(
                variable=["t"],
                level=[500, 700],
                date=20180801,
                time=1200,
            ),
            [
                ["t", 700, 20180801, 1200],
                ["t", 500, 20180801, 1200],
            ],
            ["variable", "level", "date", "time"],
        ),
    ],
)
def test_netcdf_sel_single_file_1(mode, params, expected_meta, metadata_keys):
    f = load_nc_or_xr_source(earthkit_examples_file("tuv_pl.nc"), mode)

    g = f.sel(**params)
    assert len(g) == len(expected_meta)
    if len(expected_meta) > 0:
        keys = list(params.keys())
        if metadata_keys:
            keys = metadata_keys

        assert g.metadata(keys) == expected_meta
    return


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
from earthkit.data.translators import Translator


class StrTranslator(Translator):
    def __init__(self, data, *args, **kwargs):
        self.data = data.to_string(*args, **kwargs)


def translator(data, cls, *args, **kwargs):
    if cls in [str, "string"]:
        return StrTranslator(data, *args, **kwargs)

    return None



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


def metadata_argument(*args, namespace=None, astype=None):
    """Helps parsing the input arguments for the metadata methods"""
    key = []
    key_arg_type = None
    if len(args) == 1 and isinstance(args[0], str):
        key_arg_type = str
    elif len(args) >= 1:
        key_arg_type = tuple
        for k in args:
            if isinstance(k, list):
                key_arg_type = list

    for k in args:
        if isinstance(k, str):
            key.append(k)
        elif isinstance(k, (list, tuple)):
            key.extend(k)
        else:
            raise ValueError(f"metadata: invalid key argument={k}")

    if key:
        if namespace is not None:
            if not isinstance(namespace, str):
                raise ValueError(f"metadata: namespace={namespace} must be a str when key specified")

        if isinstance(astype, (list, tuple)):
            if len(astype) != len(key):
                if len(astype) == 1:
                    astype = [astype[0]] * len(key)
                else:
                    raise ValueError("metadata: astype must have the same number of items as key")
        else:
            astype = [astype] * len(key)

    if namespace is None:
        namespace = []
    elif isinstance(namespace, str) and namespace == "":
        namespace = []
    elif isinstance(namespace, str) or namespace == all:
        namespace = [namespace]
    elif namespace == [""]:
        namespace = []

    return (key, namespace, astype, key_arg_type)



import earthkit.data

# Reset a named config option to its default value
earthkit.data.config.reset("user-cache-directory")

# Reset all the config options to their default values
earthkit.data.config.reset()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

import earthkit.data
from earthkit.data.testing import earthkit_test_data_file


def test_empty_file_reader():
    with pytest.raises(Exception):
        earthkit.data.from_source("file", earthkit_test_data_file("empty_file.grib"))


def test_nonexisting_file_reader():
    with pytest.raises(FileNotFoundError):
        earthkit.data.from_source("file", "__nonexistingfile__")


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


PROJ_PARAMS_TO_KWARGS = {
    "lon_0": "central_longitude",
    "lat_0": "central_latitude",
    "lat_ts": "true_scale_latitude",
    "x_0": "false_easting",
    "y_0": "false_northing",
    "k": "scale_factor",
    "zone": "zone",
    "proj": "proj_name",
}

STANDARD_PARALLELS = ["lat_1", "lat_2"]

PROJ_PARAMS_TO_GLOBE_KWARGS = {
    "a": "semimajor_axis",
    "b": "semiminor_axis",
    "datum": "datum",
    "ellps": "ellipse",
    "f": "flattening",
    "rf": "inverse_flattening",
}


def to_dict(proj_string):
    proj_params = {
        k.lstrip("+"): v for k, v in (p.split("=") if "=" in p else (p, None) for p in proj_string.split())
    }

    for key, value in proj_params.items():
        try:
            proj_params[key] = float(value)
        except (TypeError, ValueError):
            continue

    return proj_params


def to_projection_kwargs(proj_params):
    if isinstance(proj_params, str):
        proj_params = to_dict(proj_params)
    proj_params.pop("proj", None)  # remove the projection name

    globe_params = {
        PROJ_PARAMS_TO_GLOBE_KWARGS[param]: proj_params.pop(param)
        for param in PROJ_PARAMS_TO_GLOBE_KWARGS
        if param in proj_params
    }

    kwargs = {PROJ_PARAMS_TO_KWARGS[k]: v for k, v in proj_params.items() if k in PROJ_PARAMS_TO_KWARGS}
    kwargs["globe"] = globe_params

    standard_parallels = tuple(proj_params[param] for param in STANDARD_PARALLELS if param in proj_params)
    if standard_parallels:
        kwargs["standard_parallels"] = standard_parallels

    return kwargs



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from forcings_fixtures import load_forcings_fs  # noqa: E402


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
def test_forcings_single_index_bad(input_data):
    ds, _ = load_forcings_fs(input_data=input_data)
    idx = len(ds) + 10
    with pytest.raises(IndexError):
        ds[idx]


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
@pytest.mark.parametrize("index", [0, 2, 95, -1, -96])
def test_forcings_single_index(input_data, index):
    ds, md = load_forcings_fs(input_data=input_data)
    num = len(ds)
    r = ds[index]

    ref_md = md[index]
    assert r.metadata(["valid_datetime", "param"]) == ref_md

    v = r.values
    assert v.shape == (209,)

    # check the original fieldlist
    assert len(ds) == num


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
@pytest.mark.parametrize(
    "indexes",
    [
        slice(0, 4),
        slice(None, 4),
        slice(2, 9, 2),
        slice(8, 1, -2),
        slice(14, 18),
        slice(91, None),
    ],
)
def test_forcings_slice(input_data, indexes):
    ds, md = load_forcings_fs(input_data=input_data)
    num = len(ds)
    r = ds[indexes]

    ref_md = md[indexes]
    ref_num = len(ref_md)
    assert len(r) == ref_num
    assert r.metadata(["valid_datetime", "param"]) == ref_md

    v = r.values
    assert v.shape == (ref_num, 209)

    # check the original fieldlist
    assert len(ds) == num


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
@pytest.mark.parametrize(
    "indexes1,indexes2",
    [
        (np.array([1, 16, 5, 9]), np.array([1, 3])),
        ([1, 16, 5, 9], [1, 3]),
        ((1, 16, 5, 9), (1, 3)),
    ],
)
def test_forcings_array_indexing(input_data, indexes1, indexes2):
    ds, md = load_forcings_fs(input_data=input_data)

    # first subset
    r = ds[indexes1]
    ref_md = [md[i] for i in indexes1]
    ref_num = len(ref_md)
    assert len(r) == ref_num
    assert r.metadata(["valid_datetime", "param"]) == ref_md

    # subsetting the first subset
    r1 = r[indexes2]
    ref_md = [ref_md[i] for i in indexes2]
    ref_num = len(ref_md)
    assert len(r1) == ref_num
    assert r1.metadata(["valid_datetime", "param"]) == ref_md


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
@pytest.mark.skip(reason="Index range checking disabled")
@pytest.mark.parametrize(
    "indexes",
    [
        (np.array([1, 96, 5, 9])),
        ([1, 16, 5, 9], [1, 3]),
        ((1, 16, 5, 9), (1, 3)),
    ],
)
def test_forcings_array_indexing_bad(input_data, indexes):
    ds, _ = load_forcings_fs(input_data=input_data)
    with pytest.raises(IndexError):
        ds[indexes]


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
def test_forcings_fieldlist_iterator(input_data):
    ds, md = load_forcings_fs(input_data=input_data)
    # sn = ds.metadata(["valid_datetime", "param"])
    sn = md
    assert len(sn) == len(ds)
    iter_sn = [f.metadata(["valid_datetime", "param"]) for f in ds]
    assert iter_sn == sn
    # repeated iteration
    iter_sn = [f.metadata(["valid_datetime", "param"]) for f in ds]
    assert iter_sn == sn


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging

from earthkit.data import from_object
from earthkit.data import wrappers
from earthkit.data.wrappers import string as strwrapper

LOG = logging.getLogger(__name__)


def test_string_wrapper():
    _wrapper = strwrapper.wrapper(" ")
    assert isinstance(_wrapper, strwrapper.StrWrapper)
    _wrapper = wrappers.get_wrapper(" ")
    assert isinstance(_wrapper, strwrapper.StrWrapper)
    _wrapper = from_object(" ")
    assert isinstance(_wrapper, strwrapper.StrWrapper)


def test_datetime_string():
    import datetime

    _datetime = datetime.datetime(2000, 1, 1)
    date_formats = ["%Y-%m-%d"]
    # Currently not accepted: "%Y/%m/%d", "%d/%m/%Y", "%d-%m-%Y"
    for d_format in date_formats:
        _wrapper = strwrapper.wrapper(_datetime.strftime(d_format))
        assert _datetime == _wrapper.datetime()


def test_datetime_list_string():
    import datetime

    _datetime_list = [datetime.datetime(2000, 1, i) for i in range(1, 30)]
    date_formats = ["%Y-%m-%d"]
    # Currently not accepted: "%Y/%m/%d", "%d/%m/%Y", "%d-%m-%Y"
    for d_format in date_formats:
        datetime_list_string = "/to/".join([_datetime_list[i].strftime(d_format) for i in (0, -1)])
        _wrapper = strwrapper.wrapper(datetime_list_string)
        assert _datetime_list == _wrapper.to_datetime_list()


def test_datetime_list_string_by():
    import datetime

    _datetime_list = [datetime.datetime(2000, 1, i) for i in range(1, 30, 2)]
    date_formats = ["%Y-%m-%d"]
    # Currently not accepted: "%Y/%m/%d", "%d/%m/%Y", "%d-%m-%Y"
    for d_format in date_formats:
        datetime_list_string = "/to/".join([_datetime_list[i].strftime(d_format) for i in (0, -1)])
        _wrapper = strwrapper.wrapper(datetime_list_string + "/by/2")
        assert _datetime_list == _wrapper.to_datetime_list()


def test_boundingbox_string():
    import datetime

    _datetime_list = [datetime.datetime(2000, 1, i) for i in range(1, 30, 2)]
    date_formats = ["%Y-%m-%d"]
    # Currently not accepted: "%Y/%m/%d", "%d/%m/%Y", "%d-%m-%Y"
    for d_format in date_formats:
        datetime_list_string = "/to/".join([_datetime_list[i].strftime(d_format) for i in (0, -1)])
        _wrapper = strwrapper.wrapper(datetime_list_string + "/by/2")
        assert _datetime_list == _wrapper.to_datetime_list()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import IN_GITHUB
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import write_to_file


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_netcdf_fieldlist_save(write_method):
    ds = from_source("file", earthkit_examples_file("test.nc"))

    # the file must be saved without loading the fields
    # assert ds._reader._fields is None
    assert ds._fields is None

    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds)
        # assert ds._reader._fields is None
        assert ds._fields is None
        assert os.path.exists(tmp)
        r_tmp = from_source("file", tmp)
        assert len(r_tmp) == 2


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_netcdf_fieldlist_subset_save_1(write_method):
    ds = from_source("file", earthkit_examples_file("test.nc"))
    assert len(ds) == 2
    r = ds[1]

    with temp_file() as tmp:
        with pytest.raises(NotImplementedError):
            write_to_file(write_method, tmp, r)


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_netcdf_fieldlist_subset_save_2(write_method):
    ds = from_source("file", earthkit_examples_file("tuv_pl.nc"))
    assert len(ds) == 18
    r = ds[1:4]

    with temp_file() as tmp:
        with pytest.raises(NotImplementedError):
            write_to_file(write_method, tmp, r)


@pytest.mark.skipif(IN_GITHUB, reason="Some runners crash in Xarray")
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_netcdf_fieldlist_multi_subset_save(write_method):
    ds1 = from_source("file", earthkit_examples_file("test.nc"))
    ds2 = from_source("file", earthkit_examples_file("tuv_pl.nc"))

    ds = ds1 + ds2
    assert len(ds) == 20

    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds)
        assert os.path.exists(tmp)
        r_tmp = from_source("file", tmp)
        assert len(r_tmp) == 20


@pytest.mark.skipif(IN_GITHUB, reason="Some runners crash in Xarray")
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_netcdf_fieldlist_multi_subset_save_bad(write_method):
    ds1 = from_source("file", earthkit_examples_file("test.nc"))
    ds2 = from_source("file", earthkit_examples_file("tuv_pl.nc"))

    ds = ds1 + ds2[1:5]
    assert len(ds) == 6

    with temp_file() as tmp:
        with pytest.raises(NotImplementedError):
            write_to_file(write_method, tmp, ds)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.core.fieldlist import FieldList

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from array_fl_fixtures import check_array_fl  # noqa: E402
from array_fl_fixtures import check_save_to_disk  # noqa: E402
from array_fl_fixtures import load_array_fl  # noqa: E402


@pytest.mark.parametrize("mode", ["oper", "multi"])
def test_array_fl_grib_concat_2a(mode):
    ds1, ds2, md = load_array_fl(2)

    if mode == "oper":
        ds = ds1 + ds2
    else:
        ds = from_source("multi", ds1, ds2)

    check_array_fl(ds, [ds1, ds2], md)
    check_save_to_disk(ds, 8, md)


def test_array_fl_grib_concat_2b():
    ds1, ds2, md = load_array_fl(2)
    ds1_ori = ds1
    ds1 += ds2

    check_array_fl(ds1, [ds1_ori, ds2], md)
    check_save_to_disk(ds1, 8, md)


@pytest.mark.parametrize("mode", ["oper", "multi"])
def test_array_fl_grib_concat_3a(mode):
    ds1, ds2, ds3, md = load_array_fl(3)

    if mode == "oper":
        ds = ds1 + ds2
        ds = ds + ds3
    else:
        ds = from_source("multi", ds1, ds2)
        ds = from_source("multi", ds, ds3)

    check_array_fl(ds, [ds1, ds2, ds3], md)
    check_save_to_disk(ds, 26, md)


@pytest.mark.parametrize("mode", ["oper", "multi"])
def test_array_fl_grib_concat_3b(mode):
    ds1, ds2, ds3, md = load_array_fl(3)

    if mode == "oper":
        ds = ds1 + ds2 + ds3
    else:
        ds = from_source("multi", ds1, ds2, ds3)

    check_array_fl(ds, [ds1, ds2, ds3], md)
    check_save_to_disk(ds, 26, md)


def test_array_fl_grib_from_empty_1():
    ds_e = FieldList()
    ds, md = load_array_fl(1)
    ds1 = ds_e + ds
    assert id(ds1) == id(ds)
    assert len(ds1) == 2
    assert ds1.metadata("param") == md
    check_save_to_disk(ds1, 2, md)


def test_array_fl_grib_from_empty_2():
    ds_e = FieldList()
    ds, md = load_array_fl(1)
    ds1 = ds + ds_e
    assert id(ds1) == id(ds)
    assert len(ds1) == 2
    assert ds1.metadata("param") == md
    check_save_to_disk(ds1, 2, md)


def test_array_fl_grib_from_empty_3():
    ds_e = FieldList()
    ds1, ds2, md = load_array_fl(2)
    ds = ds_e + ds1 + ds2

    check_array_fl(ds, [ds1, ds2], md)
    check_save_to_disk(ds, 8, md)


def test_array_fl_grib_from_empty_4():
    ds = FieldList()
    ds1, md = load_array_fl(1)
    ds += ds1
    assert id(ds) == id(ds1)
    assert len(ds) == 2
    assert ds.metadata("param") == md
    check_save_to_disk(ds, 2, md)


def test_array_fl_grib_from_empty_5():
    ds = FieldList()
    ds1, ds2, md = load_array_fl(2)
    ds += ds1 + ds2

    check_array_fl(ds, [ds1, ds2], md)
    check_save_to_disk(ds, 8, md)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import config
from earthkit.data import from_source
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import earthkit_examples_file


@pytest.mark.parametrize(
    "file_path,expected_type,expected_len",
    [
        ("test.grib", "grib", 2),
        ("temp_10.bufr", "bufr", 10),
        ("test.odb", "odb", None),
    ],
)
def test_reader_padding_bytes(file_path, expected_type, expected_len):
    path = earthkit_examples_file(file_path)
    tmp = temp_file()

    # note: "reader-type-check-bytes" is 64 by default

    # prepend padding
    with open(tmp.path, "wb") as f_tmp, open(path, "rb") as f:
        b = bytes([0 for _ in range(20)])
        f_tmp.write(b)
        f_tmp.write(f.read())

    ds = from_source("file", tmp.path)
    if hasattr(ds, "_reader"):
        assert expected_type in str(type(ds._reader)).lower()
    else:
        assert expected_type in str(type(ds)).lower()
    if expected_len is not None:
        assert len(ds) == expected_len

    # prepend padding
    tmp = temp_file()
    with open(tmp.path, "wb") as f_tmp, open(path, "rb") as f:
        b = bytes([0 for _ in range(80)])
        f_tmp.write(b)
        f_tmp.write(f.read())

    ds = from_source("file", tmp.path)
    assert "file.File" in str(type(ds))
    assert "unknown" in str(type(ds._reader)).lower()

    with config.temporary("reader-type-check-bytes", 100):
        ds = from_source("file", tmp.path)
        if hasattr(ds, "_reader"):
            assert expected_type in str(type(ds._reader)).lower()
        else:
            assert expected_type in str(type(ds)).lower()
        if expected_len is not None:
            assert len(ds) == expected_len


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.testing import NO_FDB
from earthkit.data.testing import earthkit_test_data_file

TEST_GRIB_REQUEST = {
    "class": "od",
    "expver": "0001",
    "stream": "oper",
    "date": [20240603, 20240604],
    "time": [0, 1200],
    "domain": "g",
    "type": "fc",
    "levtype": "pl",
    "levelist": [500, 700],
    "step": [0, 6],
    "param": [130, 157],
}


def make_fdb_config(path):
    fdb_schema = earthkit_test_data_file("fdb_schema.txt")
    fdb_dir = path
    os.makedirs(fdb_dir, exist_ok=True)
    config = {
        "type": "local",
        "engine": "toc",
        "schema": fdb_schema,
        "spaces": [{"handler": "Default", "roots": [{"path": fdb_dir}]}],
    }
    return config


def make_fdb(path):
    ds = from_source("sample", "pl.grib")
    config = make_fdb_config(path)
    ds.to_target("fdb", config=config)
    return ds, config


@pytest.mark.skipif(NO_FDB, reason="No access to FDB")
@pytest.mark.cache
def test_lazy_fdb():
    with temp_directory() as tmpdir:
        ds, config = make_fdb(os.path.join(tmpdir, "_fdb"))

        ds = from_source("fdb", TEST_GRIB_REQUEST, config=config, stream=False, lazy=True)
        assert len(ds) == 32

        assert ds[0].shape == (19, 36)
        assert ds[1].shape == (19, 36)
        assert ds[0].metadata(["shortName", "param", "units", "cfName"]) == ["t", "t", "K", "air_temperature"]
        assert ds[1].metadata(["shortName", "param", "units", "cfName"]) == [
            "r",
            "r",
            "%",
            "relative_humidity",
        ]

        assert not hasattr(ds, "path")
        assert not hasattr(ds[0], "path")

        a = ds.to_xarray(time_dim_mode="forecast")
        assert a["t"].values.shape == (4, 2, 2, 19, 36)
        assert a["r"].values.shape == (4, 2, 2, 19, 36)

        m = a.mean("step").load()
        assert m["t"].values.shape == (4, 2, 19, 36)
        assert m["r"].values.shape == (4, 2, 19, 36)
        assert np.allclose(m["r"].values.flatten()[85:87], [47.66908598, 53.43959379])
        assert np.allclose(m["t"].values.flatten()[85:87], [253.22625732, 252.78778076])



import earthkit.data

print(earthkit.data.config.get("url-download-timeout"))

with earthkit.data.config.temporary():
    earthkit.data.config.set("url-download-timeout", 5)
    print(earthkit.data.config.get("url-download-timeout"))

# Temporary config can also be created with arguments:
with earthkit.data.config.temporary("url-download-timeout", 11):
    print(earthkit.data.config.get("url-download-timeout"))



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import os
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from grib_fixtures import FL_ARRAYS  # noqa: E402
from grib_fixtures import load_grib_data  # noqa: E402


@pytest.mark.parametrize("fl_type", FL_ARRAYS)
@pytest.mark.parametrize("group", ["param"])
def test_grib_group_by(fl_type, group):
    ds, array_backend = load_grib_data("test6.grib", fl_type)

    ref = [
        [("t", 1000), ("t", 850)],
        [("u", 1000), ("u", 850)],
        [("v", 1000), ("v", 850)],
    ]
    cnt = 0
    for i, f in enumerate(ds.group_by(group)):
        assert len(f) == 2
        assert f.metadata(("param", "level")) == ref[i]
        afl = f.to_fieldlist(array_backend=array_backend.name)
        assert afl is not f
        assert len(afl) == 2
        cnt += len(f)

    assert cnt == len(ds)


@pytest.mark.parametrize("fl_type", FL_ARRAYS)
@pytest.mark.parametrize("group", ["level", ["level", "gridType"]])
def test_grib_multi_group_by(fl_type, group):
    ds, _ = load_grib_data(["test4.grib", "test6.grib"], fl_type)

    ref = [
        [("t", 500), ("z", 500)],
        [("t", 850), ("z", 850), ("t", 850), ("u", 850), ("v", 850)],
        [("t", 1000), ("u", 1000), ("v", 1000)],
    ]
    cnt = 0
    for i, f in enumerate(ds.group_by(group)):
        assert f.metadata(("param", "level")) == ref[i]
        cnt += len(f)

    assert cnt == len(ds)


@pytest.mark.parametrize(
    "_kwargs,expected_meta",
    [
        ({"n": 1}, [["t"], ["u"], ["v"], ["t"], ["u"], ["v"]]),
        ({"n": 3}, [["t", "u", "v"], ["t", "u", "v"]]),
        ({"n": 4}, [["t", "u", "v", "t"], ["u", "v"]]),
    ],
)
def test_grib_batched(_kwargs, expected_meta):
    ds = from_source("file", earthkit_examples_file("test6.grib"))

    cnt = 0
    for i, f in enumerate(ds.batched(_kwargs["n"])):
        assert len(f) == len(expected_meta[i])
        f.metadata("param") == expected_meta[i]
        cnt += len(f)

    assert cnt == len(ds)


@pytest.mark.parametrize(
    "_kwargs,expected_meta",
    [
        ({"n": 1}, [["2t"], ["msl"], ["t"], ["z"], ["t"], ["z"]]),
        ({"n": 2}, [["2t", "msl"], ["t", "z"], ["t", "z"]]),
        ({"n": 3}, [["2t", "msl", "t"], ["z", "t", "z"]]),
        ({"n": 4}, [["2t", "msl", "t", "z"], ["t", "z"]]),
    ],
)
def test_grib_multi_batched(_kwargs, expected_meta):
    ds = from_source(
        "file",
        [earthkit_examples_file("test.grib"), earthkit_examples_file("test4.grib")],
    )

    cnt = 0
    n = _kwargs["n"]
    for i, f in enumerate(ds.batched(n)):
        assert len(f) == len(expected_meta[i])
        f.metadata("param") == expected_meta[i]
        cnt += len(f)

    assert cnt == len(ds)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import copy
import logging
from functools import cached_property
from math import prod

import numpy as np

from earthkit.data.core.geography import Geography
from earthkit.data.core.metadata import Metadata
from earthkit.data.core.metadata import MetadataAccessor
from earthkit.data.utils.bbox import BoundingBox
from earthkit.data.utils.dates import datetime_from_grib
from earthkit.data.utils.dates import to_datetime
from earthkit.data.utils.dates import to_timedelta
from earthkit.data.utils.projections import Projection

LOG = logging.getLogger(__name__)


def uniform_resolution(vals):
    if len(vals) > 1:
        delta = np.diff(vals)
        if np.allclose(delta, delta[0]):
            return delta[0]
    return None


def make_geography(metadata, values_shape=None):
    lat = metadata.get("latitudes", None)
    lon = metadata.get("longitudes", None)

    values_size = prod(values_shape) if values_shape else None

    distinct = False
    if lat is None or lon is None:
        lat = metadata.get("distinctLatitudes", None)
        lon = metadata.get("distinctLongitudes", None)

        # it is possible to have no geography at all.
        if lat is None and lon is None:
            return NoGeography(values_shape)

        if values_shape is None:
            if lat is None:
                raise ValueError("No latitudes or distinctLatitudes found")
            if lon is None:
                raise ValueError("No longitudes or distinctLongitudes found")

            lat = np.asarray(lat, dtype=float)
            lon = np.asarray(lon, dtype=float)

            if len(lat.shape) != 1:
                raise ValueError(f"distinct latitudes must be 1D array! shape={lat.shape} unsupported")
            if len(lon.shape) != 1:
                raise ValueError(f"distinctLongitudes must be 1D array! shape={lon.shape} unsupported")
            distinct = True
        else:
            if lat is not None and lon is not None:
                lat = np.asarray(lat, dtype=float)
                lon = np.asarray(lon, dtype=float)

                if len(lat.shape) != 1:
                    raise ValueError(f"distinct latitudes must be 1D array! shape={lat.shape} unsupported")
                if len(lon.shape) != 1:
                    raise ValueError(f"distinctLongitudes must be 1D array! shape={lon.shape} unsupported")
                if lat.size * lon.size != values_size:
                    raise ValueError(
                        (
                            "Distinct latitudes and longitudes do not match number of values. "
                            f"Expected number=({lat.size * lon.size}), got={values_size}"
                        )
                    )
                distinct = True

            else:
                lat = None
                lon = None
    else:
        lat = np.asarray(lat, dtype=float)
        lon = np.asarray(lon, dtype=float)
        if values_size is not None:
            if lat.size * lon.size == values_size:
                if len(lat.shape) != 1:
                    raise ValueError(
                        f"latitudes must be a 1D array when holding distinct values! shape={lat.shape} unsupported"
                    )
                if len(lon.shape) != 1:
                    raise ValueError(
                        f"longitudes must be a 1D array when holding distinct values! shape={lon.shape} unsupported"
                    )
                distinct = True

    assert lat is not None and lon is not None

    if distinct:
        dx = uniform_resolution(lon)
        dy = uniform_resolution(lat)

        if dx is not None and dy is not None:
            # metadata["DxInDegrees"] = dx
            # metadata["DyInDegrees"] = dy
            return RegularDistinctLLGeography(metadata)
        else:
            return DistinctLLGeography(metadata)
    else:
        if lat.shape != lon.shape:
            raise ValueError(f"latitudes and longitudes must have the same shape. {lat.shape} != {lon.shape}")

        if values_shape is not None:
            if lat.size == values_size:
                if values_shape is not None:
                    if lat.shape != values_shape:
                        shape = lat.shape if lat.ndim > len(values_shape) else values_shape
                    else:
                        shape = lat.shape
                else:
                    shape = lat.shape

                return UserGeography(metadata, shape=shape)

            else:
                raise ValueError(
                    (
                        "latitudes and longitudes do not match number of values. "
                        f"Expected number=({lat.size * lon.size}), got={values_size}"
                    )
                )
        else:
            shape = lat.shape
            return UserGeography(metadata, shape=shape)


class NoGeography(Geography):
    def __init__(self, shape):
        self._shape = shape

    def latitudes(self, dtype=None):
        return None

    def longitudes(self, dtype=None):
        return None

    def x(self, dtype=None):
        raise NotImplementedError("x is not implemented for this geography")

    def y(self, dtype=None):
        raise NotImplementedError("y is not implemented for this geography")

    def shape(self):
        return self._shape

    def _unique_grid_id(self):
        return self.shape()

    def projection(self):
        return None

    def bounding_box(self):
        return None

    def gridspec(self):
        return None

    def resolution(self):
        return None
        # raise NotImplementedError("resolution is not implemented for this geography")

    def mars_area(self):
        return None

    def mars_grid(self):
        raise NotImplementedError("mars_grid is not implemented for this geography")

    def grid_type(self):
        return "none"


class UserGeography(Geography):
    def __init__(self, metadata, shape=None):
        self.metadata = metadata
        self._shape = shape

    def latitudes(self, dtype=None):
        v = self.metadata.get("latitudes")
        v = np.asarray(v)
        if dtype is None:
            return v
        else:
            return v.astype(dtype)

    def longitudes(self, dtype=None):
        v = self.metadata.get("longitudes")
        v = np.asarray(v)
        if dtype is None:
            return v
        else:
            return v.astype(dtype)

    def x(self, dtype=None):
        raise NotImplementedError("x is not implemented for this geography")

    def y(self, dtype=None):
        raise NotImplementedError("y is not implemented for this geography")

    def shape(self):
        if self._shape is not None:
            return self._shape
        return self.latitudes().shape

    def _unique_grid_id(self):
        return self.shape()

    def north(self):
        return np.amax(self.latitudes())

    def south(self):
        return np.amin(self.latitudes())

    def west(self):
        return np.amin(self.longitudes())

    def east(self):
        return np.amax(self.longitudes())

    def projection(self):
        return Projection.from_proj_string(self.metadata.get("projTargetString", None))

    def bounding_box(self):
        return BoundingBox(
            north=self.north(),
            south=self.south(),
            west=self.west(),
            east=self.east(),
        )

    def gridspec(self):
        return None

    def resolution(self):
        return None
        # raise NotImplementedError("resolution is not implemented for this geography")

    def mars_area(self):
        return [self.north(), self.west(), self.south(), self.east()]

    def mars_grid(self):
        raise NotImplementedError("mars_grid is not implemented for this geography")

    def grid_type(self):
        return "_unstructured"


class DistinctLLGeography(UserGeography):
    def __init__(self, metadata):
        super().__init__(metadata)

    def _distinct_latitudes(self):
        v = self.metadata.get("latitudes", None)
        if v is None:
            v = self.metadata.get("distinctLatitudes", None)
        if v is None:
            raise ValueError("No latitudes found")
        return np.asarray(v)

    def _distinct_longitudes(self):
        v = self.metadata.get("longitudes", None)
        if v is None:
            v = self.metadata.get("distinctLongitudes", None)
        if v is None:
            raise ValueError("No longitudes found")
        return np.asarray(v)

    def latitudes(self, dtype=None):
        lat = self._distinct_latitudes()
        n_lon = len(self._distinct_longitudes())
        v = np.repeat(lat[:, np.newaxis], n_lon, axis=1)

        if dtype is None:
            return v
        else:
            return v.astype(dtype)

    def longitudes(self, dtype=None):
        lon = self._distinct_longitudes()
        n_lat = len(self._distinct_latitudes())
        v = np.repeat(lon[np.newaxis, :], n_lat, axis=0)

        if dtype is None:
            return v
        else:
            return v.astype(dtype)

    def shape(self):
        Nj = len(self._distinct_latitudes())
        Ni = len(self._distinct_longitudes())
        return (Nj, Ni)

    def grid_type(self):
        return "_distinct_ll"


class RegularDistinctLLGeography(DistinctLLGeography):
    def dx(self):
        x = self.metadata.get("DxInDegrees", None)
        if x is None:
            lon = self._distinct_longitudes()
            x = lon[1] - lon[0]
        x = abs(round(x * 1_000_000) / 1_000_000)
        return x

    def dy(self):
        y = self.metadata.get("DyInDegrees", None)
        if y is None:
            lat = self._distinct_latitudes()
            y = lat[0] - lat[1]
        y = abs(round(y * 1_000_000) / 1_000_000)
        return y

    def resolution(self):
        x = self.dx()
        y = self.dy()
        if x == y:
            return x

    def mars_grid(self):
        return [self.dx(), self.dy()]

    def grid_type(self):
        return "_regular_ll"


class UserMetadata(Metadata):
    ALIASES = [
        ("dataDate", "date"),
        ("dataTime", "time"),
        ("forecast_reference_time", "base_datetime"),
        ("level", "levelist"),
        ("step", "endStep", "stepRange"),
        ("param", "shortName"),
    ]

    ACCESSORS = {
        "base_datetime": "base_datetime",
        "valid_datetime": "valid_datetime",
        "step_timedelta": "step_timedelta",
        "param_level": "param_level",
        "_grid_type": "gridType",
    }

    LS_KEYS = ["param", "level", "base_datetime", "valid_datetime", "step", "number"]

    def __init__(self, d, shape=None, **kwargs):
        self._data = d
        self._shape = shape

    def __len__(self):
        return len(self._data)

    def __contains__(self, key):
        return key in self._data

    def __getitem__(self, key):
        return self.get(key, raise_on_missing=True)

    def __iter__(self):
        return iter(self._keys())

    def keys(self):
        return self._data.keys()

    def items(self):
        return self._data.items()

    @MetadataAccessor(ACCESSORS, ALIASES)
    def get(self, key, default=None, *, astype=None, raise_on_missing=False):
        def _key_name(key):
            if key in self._data:
                return key
            else:
                for group in self.ALIASES:
                    if key in group:
                        for k in group:
                            if k in self._data:
                                return k

        key_n = _key_name(key)
        if key_n is None:
            if raise_on_missing:
                raise KeyError(f"Key={key} not found")
            return default

        v = self._data[key_n]
        if astype is None:
            return v
        else:
            return astype(v)

    def datetime(self):
        return {
            "base_time": self.base_datetime(),
            "valid_time": self.valid_datetime(),
        }

    def base_datetime(self):
        v = self._get_one(["base_datetime", "forecast_reference_time"])
        if v is not None:
            v = to_datetime(v)
            return v

        v = self._datetime("hdate", "time")
        if v is not None:
            return v

        v = self._datetime("date", "time")
        if v is not None:
            return v

        v = self.step_timedelta()
        if v is not None:
            return self.valid_datetime() - v

    def valid_datetime(self):
        if "valid_datetime" in self._data:
            v = self._data["valid_datetime"]
            v = to_datetime(v)
            return v

        base_dt = self.base_datetime()
        if base_dt is not None:
            td = self.step_timedelta()
            if td is not None:
                return base_dt + td
            return base_dt

    def step_timedelta(self):
        if "step_timedelta" in self._data:
            return self._data["step_timedelta"]
        v = self._get_one(["endStep", "step"])
        if v is not None:
            return to_timedelta(v)

    def _datetime(self, date_key, time_key):
        date = self.get(date_key, None)
        if date is not None:
            time = self.get(time_key, None)
            if time is not None:
                return datetime_from_grib(date, time)
        return None

    def param_level(self):
        return f"{self.get('param')}{self.get('level', default='')}"

    def _grid_type(self):
        if "gridType" in self._data:
            return self._data["gridType"]
        return self.geography.grid_type()

    def _get_one(self, keys):
        for k in keys:
            if k in self._data:
                return self._data[k]

    @cached_property
    def geography(self):
        return make_geography(self, values_shape=self._shape)

    def override(self, *args, **kwargs):
        r"""Create a new metadata object by cloning the underlying metadata and setting the keys in it.

        Parameters
        ----------
        *args: tuple
            Positional arguments. When present must be a dict with the keys to set in
            the new metadata.
        **kwargs: dict, optional
            Other keyword arguments specifying the metadata keys to set.

        Returns
        -------
        :class:`UserMetadata`
            The new metadata object. A copy of the original metadata with the keys set in it.
        """
        d = dict(*args, **kwargs)
        existing = copy.deepcopy(self._data)
        existing.update(d)

        return UserMetadata(existing, shape=copy.deepcopy(self._shape))

    def namespaces(self):
        return []

    def as_namespace(self, namespace=None):
        return {}

    def dump(self, **kwargs):
        r"""Generate dump with all the metadata keys.

        In a Jupyter notebook it is represented as a tabbed interface.

        Parameters
        ----------
        **kwargs: dict, optional
            Other keyword arguments used for testing only

        Returns
        -------
        NamespaceDump
            Dict-like object with one item per namespace. In a Jupyter notebook represented
            as a tabbed interface to browse the dump contents.

        Examples
        --------
        :ref:`/examples/grib_metadata.ipynb`

        """
        from earthkit.data.utils.summary import format_namespace_dump

        r = [
            {
                "title": "metadata",
                "data": self._data,
            }
        ]
        return format_namespace_dump(r, selected="parameter", details=self.__class__.__name__, **kwargs)

    def ls_keys(self):
        return self.LS_KEYS

    def describe_keys(self):
        return []

    def index_keys(self):
        return None

    def data_format(self):
        return "dict"

    def _hide_internal_keys(self):
        return self



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import eccodes

from earthkit.data.indexing.fieldlist import ClonedFieldCore
from earthkit.data.indexing.fieldlist import SimpleFieldList
from earthkit.data.readers import Reader
from earthkit.data.readers.grib.codes import GribCodesHandle
from earthkit.data.readers.grib.codes import GribField

LOG = logging.getLogger(__name__)


def get_use_grib_metadata_cache():
    from earthkit.data.core.config import CONFIG

    return CONFIG.get("use-grib-metadata-cache")


class GribMemoryReader(Reader):
    def __init__(self, use_grib_metadata_cache=None, **kwargs):
        self._peeked = None
        self.consumed_ = False
        self._use_metadata_cache = (
            use_grib_metadata_cache if use_grib_metadata_cache is not None else get_use_grib_metadata_cache()
        )

    def __iter__(self):
        return self

    def __next__(self):
        if self._peeked is not None:
            msg = self._peeked
            self._peeked = None
            return msg
        handle = self._next_handle()
        msg = self._message_from_handle(handle)
        if handle is not None:
            return msg
        self.consumed_ = True
        raise StopIteration

    def _next_handle(self):
        raise NotImplementedError

    def _message_from_handle(self, handle):
        if handle is not None:
            return GribFieldInMemory(
                GribCodesHandle(handle, None, None), use_metadata_cache=self._use_metadata_cache
            )

    def batched(self, n):
        from earthkit.data.utils.batch import batched

        return batched(self, n, create=self.to_fieldlist)

    def group_by(self, *args, **kwargs):
        from earthkit.data.utils.batch import group_by

        return group_by(self, *args, create=self.to_fieldlist, sort=False)

    def to_fieldlist(self, fields):
        return GribFieldListInMemory.from_fields(fields)


class GribFileMemoryReader(GribMemoryReader):
    def __init__(self, path, **kwargs):
        super().__init__(**kwargs)
        self.fp = open(path, "rb")

    def __del__(self):
        self.fp.close()

    def _next_handle(self):
        return eccodes.codes_new_from_file(self.fp, eccodes.CODES_PRODUCT_GRIB)


class GribMessageMemoryReader(GribMemoryReader):
    def __init__(self, buf, **kwargs):
        super().__init__(**kwargs)
        self.buf = buf

    def __del__(self):
        self.buf = None

    def _next_handle(self):
        if self.buf is None:
            return None
        handle = eccodes.codes_new_from_message(self.buf)
        self.buf = None
        return handle


class GribStreamReader(GribMemoryReader):
    """Wrapper around eccodes.Streamreader. The problem is that when iterating via
    the StreamReader it returns an eccodes.GRIBMessage that releases the handle when deleted.
    However, the handle has to be managed by earthkit-data so we access it directly
    using _next_handle
    """

    def __init__(self, stream, **kwargs):
        super().__init__(**kwargs)
        self._stream = stream
        self._reader = eccodes.StreamReader(stream)

    def __del__(self):
        try:
            self._stream.close()
        except Exception:
            pass

    def _next_handle(self):
        try:
            return self._reader._next_handle()
        except Exception:
            self._stream.close()
            raise

    def mutate(self):
        return self

    def mutate_source(self):
        return self


class GribFieldInMemory(GribField):
    """Represents a GRIB message in memory"""

    def __init__(self, handle, use_metadata_cache=False):
        super().__init__(None, None, None, use_metadata_cache=use_metadata_cache)
        self._handle = handle

    @GribField.handle.getter
    def handle(self):
        return self._handle

    @GribField.handle.getter
    def offset(self):
        return None

    # @cached_property
    # def _metadata(self):
    #     return GribFieldMetadata(self)

    @staticmethod
    def to_fieldlist(fields):
        return GribFieldListInMemory.from_fields(fields)

    @staticmethod
    def from_buffer(buf):
        handle = eccodes.codes_new_from_message(buf)
        return GribFieldInMemory(
            GribCodesHandle(handle, None, None), use_metadata_cache=get_use_grib_metadata_cache()
        )

    def _release(self):
        self._handle = None

    def clone(self, **kwargs):
        return ClonedGribFieldInMemory(self, **kwargs)

    def __getstate__(self):
        return {"message": self.message()}

    def __setstate__(self, state):
        self.__init__(GribCodesHandle.from_message(state["message"]))


class ClonedGribFieldInMemory(ClonedFieldCore, GribFieldInMemory):
    def __init__(self, field, **kwargs):
        ClonedFieldCore.__init__(self, field, **kwargs)
        self._handle = field._handle
        GribFieldInMemory.__init__(
            self,
            field._handle,
            use_metadata_cache=field._use_metadata_cache,
        )


class GribFieldListInMemory(SimpleFieldList):
    """Represent a GRIB field list in memory loaded lazily"""

    def __init__(self, source, reader, *args, **kwargs):
        """The reader must support __next__."""
        self._reader = reader
        self._loaded = False

    def __len__(self):
        self._load()
        return super().__len__()

    def __getitem__(self, n):
        self._load()
        return super().__getitem__(n)

    def _load(self):
        # TODO: make it thread safe
        if not self._loaded:
            self.fields = [f for f in self._reader]
            self._loaded = True
            self._reader = None

    def mutate_source(self):
        return self

    @classmethod
    def merge(cls, readers):
        assert all(isinstance(s, GribFieldListInMemory) for s in readers), readers
        assert len(readers) > 1

        from itertools import chain

        return GribFieldListInMemory.from_fields(list(chain(*[f for f in readers])))

    def __getstate__(self):
        self._load()
        r = {"messages": [f.message() for f in self]}
        return r

    def __setstate__(self, state):
        fields = [GribFieldInMemory.from_buffer(m) for m in state["messages"]]
        self.__init__(None, None)
        self.fields = fields
        self._loaded = True



#!/usr/bin/env python3

# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import os
import sys
import tempfile

import numpy as np
import pytest

import earthkit.data
from earthkit.data import from_source
from earthkit.data import to_target
from earthkit.data.core.temporary import temp_directory
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import write_to_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from grib_fixtures import FL_ARRAYS  # noqa: E402
from grib_fixtures import load_grib_data  # noqa: E402

EPSILON = 1e-4


def new_grib_output_compat(*args, split_output=False, **kwargs):
    from earthkit.data.targets.file import FileTarget
    from earthkit.data.targets.file_pattern import FilePatternTarget

    if split_output:
        return FilePatternTarget(*args, encoder="grib", **kwargs)
    else:
        return FileTarget(*args, encoder="grib", **kwargs)


def new_grib_coder_compat(*args, **kwargs):
    from earthkit.data.encoders.grib import GribEncoder

    return GribEncoder(*args, **kwargs)


@pytest.mark.parametrize("fl_type", FL_ARRAYS)
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_save_when_loaded_from_file(fl_type, write_method):
    fs, _ = load_grib_data("test6.grib", fl_type)
    assert len(fs) == 6
    with temp_file() as tmp:
        write_to_file(write_method, tmp, fs)
        fs_saved = from_source("file", tmp)
        assert len(fs) == len(fs_saved)


@pytest.mark.parametrize(
    "_kwargs,expected_value",
    [({}, 16), ({"bits_per_value": 12}, 12), ({"bits_per_value": None}, 16)],
)
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_save_bits_per_value_fieldlist(_kwargs, expected_value, write_method):
    ds = from_source("file", earthkit_examples_file("test.grib"))

    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds, **_kwargs)
        ds1 = from_source("file", tmp)
        assert ds1.metadata("bitsPerValue") == [expected_value] * len(ds)


@pytest.mark.parametrize(
    "_kwargs,expected_value",
    [({}, 16), ({"bits_per_value": 12}, 12), ({"bits_per_value": None}, 16)],
)
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_save_bits_per_value_single_field(_kwargs, expected_value, write_method):
    ds = from_source("file", earthkit_examples_file("test.grib"))

    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds[0], **_kwargs)
        ds1 = from_source("file", tmp)
        assert ds1.metadata("bitsPerValue") == [expected_value]


# TODO: if we use missing_value = np.finfo(np.float32).max the test fails
@pytest.mark.parametrize("mode", ["ori", "compat", "target"])
@pytest.mark.parametrize("missing_value", [100000.0, np.finfo(np.float32).max - 1])
def test_grib_output_missing_value_1(mode, missing_value):
    fld = from_source("file", earthkit_examples_file("test.grib"))[0]

    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
        path = os.path.join(tmp, "a.grib")

        values = fld.values
        values[0] = np.nan
        assert not np.isnan(values[1])

        if mode == "ori":
            f = earthkit.data.new_grib_output(path)
            f.write(values, check_nans=True, missing_value=missing_value, template=fld)
            f.close()
        elif mode == "compat":
            f = new_grib_output_compat(path)
            f.write(values=values, check_nans=True, missing_value=missing_value, template=fld)
            f.close()
        elif mode == "target":
            to_target(
                "file",
                path,
                values=values,
                template=fld,
                check_nans=True,
                missing_value=missing_value,
            )
        ds = earthkit.data.from_source("file", path)
        assert ds[0].metadata("bitmapPresent") == 1
        assert np.isnan(ds[0].values[0])
        assert not np.isnan(values[1])


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="ignore_cleanup_errors requires Python 3.10 or later",
)
@pytest.mark.parametrize("mode", ["ori", "compat", "target"])
def test_grib_output_latlon(mode):
    data = np.random.random((181, 360))

    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
        path = os.path.join(tmp, "a.grib")

        if mode == "ori":
            f = earthkit.data.new_grib_output(path, date=20010101)
            f.write(data, param="2t")
            f.close()
        elif mode == "compat":
            f = new_grib_output_compat(path, metadata=dict(date=20010101, generatingProcessIdentifier=255))
            f.write(values=data, param="2t")
            f.close()
        elif mode == "target":
            to_target(
                "file",
                path,
                metadata=dict(date=20010101, generatingProcessIdentifier=255, param="2t"),
                values=data,
            )

        ds = earthkit.data.from_source("file", path)
        print(ds[0])

        assert ds[0].metadata("date") == 20010101
        assert ds[0].metadata("param") == "2t"
        assert ds[0].metadata("levtype") == "sfc"
        assert ds[0].metadata("edition") == 2
        assert ds[0].metadata("generatingProcessIdentifier") == 255

        assert np.allclose(ds[0].to_numpy(), data, rtol=EPSILON, atol=EPSILON)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="ignore_cleanup_errors requires Python 3.10 or later",
)
@pytest.mark.parametrize("mode", ["ori", "compat", "target"])
def test_grib_output_o96_sfc(mode):
    data = np.random.random((40320,))

    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
        path = os.path.join(tmp, "a.grib")

        if mode == "ori":
            f = earthkit.data.new_grib_output(path, date=20010101)
            f.write(data, param="2t")
            f.close()
        elif mode == "compat":
            f = new_grib_output_compat(path, metadata=dict(date=20010101, generatingProcessIdentifier=255))
            f.write(values=data, param="2t")
            f.close()
        elif mode == "target":
            to_target(
                "file",
                path,
                metadata=dict(date=20010101, generatingProcessIdentifier=255, param="2t"),
                values=data,
            )

        ds = earthkit.data.from_source("file", path)

        ref = {
            "date": 20010101,
            "param": "2t",
            "levtype": "sfc",
            "edition": 2,
            "generatingProcessIdentifier": 255,
            "gridType": "reduced_gg",
            "N": 96,
            "isOctahedral": 1,
        }

        for k, v in ref.items():
            assert ds[0].metadata(k) == v, f"{k}: {ds[0].metadata(k)}!={v}"

        assert np.allclose(ds[0].to_numpy(), data, rtol=EPSILON, atol=EPSILON)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="ignore_cleanup_errors requires Python 3.10 or later",
)
@pytest.mark.parametrize("mode", ["ori", "compat", "target"])
def test_grib_output_o160_sfc(mode):
    data = np.random.random((108160,))

    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
        path = os.path.join(tmp, "a.grib")

        if mode == "ori":
            f = earthkit.data.new_grib_output(path, date=20010101)
            f.write(data, param="2t")
            f.close()
        elif mode == "compat":
            f = new_grib_output_compat(path, metadata=dict(date=20010101, generatingProcessIdentifier=255))
            f.write(values=data, param="2t")
            f.close()
        elif mode == "target":
            to_target(
                "file",
                path,
                metadata=dict(date=20010101, generatingProcessIdentifier=255, param="2t"),
                values=data,
            )
        ds = earthkit.data.from_source("file", path)

        ref = {
            "date": 20010101,
            "param": "2t",
            "levtype": "sfc",
            "edition": 2,
            "generatingProcessIdentifier": 255,
            "gridType": "reduced_gg",
            "N": 160,
            "isOctahedral": 1,
        }

        for k, v in ref.items():
            assert ds[0].metadata(k) == v, f"{k}: {ds[0].metadata(k)}!={v}"

        assert np.allclose(ds[0].to_numpy(), data, rtol=EPSILON, atol=EPSILON)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="ignore_cleanup_errors requires Python 3.10 or later",
)
@pytest.mark.parametrize("mode", ["ori", "compat", "target"])
def test_grib_output_n96_sfc(mode):
    data = np.random.random(50662)

    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
        path = os.path.join(tmp, "a.grib")

        if mode == "ori":
            f = earthkit.data.new_grib_output(path, date=20010101)
            f.write(data, param="2t")
            f.close()
        elif mode == "compat":
            f = new_grib_output_compat(path, metadata=dict(date=20010101, generatingProcessIdentifier=255))
            f.write(values=data, param="2t")
            f.close()
        else:
            to_target(
                "file",
                path,
                metadata=dict(date=20010101, generatingProcessIdentifier=255, param="2t"),
                values=data,
            )
        ds = earthkit.data.from_source("file", path)

        ref = {
            "date": 20010101,
            "param": "2t",
            "levtype": "sfc",
            "edition": 2,
            "generatingProcessIdentifier": 255,
            "gridType": "reduced_gg",
            "N": 96,
            "isOctahedral": 0,
        }

        for k, v in ref.items():
            assert ds[0].metadata(k) == v, f"{k}: {ds[0].metadata(k)}!={v}"

        assert np.allclose(ds[0].to_numpy(), data, rtol=EPSILON, atol=EPSILON)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="ignore_cleanup_errors requires Python 3.10 or later",
)
@pytest.mark.parametrize("mode", ["ori", "compat", "target"])
def test_grib_output_mars_labeling(mode):
    data = np.random.random((40320,))

    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
        path = os.path.join(tmp, "a.grib")

        if mode == "ori":
            f = earthkit.data.new_grib_output(path, date=20010101)
            f.write(data, type="fc", expver="test", step=24, param="msl")
            f.close()
        elif mode == "compat":
            f = new_grib_output_compat(path, metadata=dict(date=20010101, generatingProcessIdentifier=255))
            f.write(values=data, type="fc", expver="test", step=24, param="msl")
            f.close()
        elif mode == "target":
            to_target(
                "file",
                path,
                metadata=dict(
                    date=20010101,
                    generatingProcessIdentifier=255,
                    type="fc",
                    expver="test",
                    step=24,
                    param="msl",
                ),
                values=data,
            )

        ds = earthkit.data.from_source("file", path)

        assert ds[0].metadata("date") == 20010101
        assert ds[0].metadata("edition") == 2
        assert ds[0].metadata("step", astype=int) == 24
        assert ds[0].metadata("expver") == "test"
        assert ds[0].metadata("levtype") == "sfc"
        assert ds[0].metadata("param") == "msl"
        assert ds[0].metadata("type") == "fc"
        assert ds[0].metadata("generatingProcessIdentifier") == 255
        assert ds[0].metadata("gridType") == "reduced_gg"
        assert ds[0].metadata("N") == 96
        assert ds[0].metadata("isOctahedral") == 1

        assert np.allclose(ds[0].to_numpy(), data, rtol=EPSILON, atol=EPSILON)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="ignore_cleanup_errors requires Python 3.10 or later",
)
@pytest.mark.parametrize("mode", ["ori", "compat", "target"])
@pytest.mark.parametrize("levtype", [{}, {"levtype": "pl"}])
def test_grib_output_o96_pl(mode, levtype):
    data = np.random.random((40320,))

    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
        path = os.path.join(tmp, "a.grib")

        if mode == "ori":
            f = earthkit.data.new_grib_output(path, date=20010101)
            _kwargs = dict(param="t", level=850)
            _kwargs.update(levtype)
            f.write(data, **_kwargs)
            f.close()
        elif mode == "compat":
            f = new_grib_output_compat(path, metadata=dict(date=20010101, generatingProcessIdentifier=255))
            _kwargs = dict(param="t", level=850)
            _kwargs.update(levtype)
            f.write(values=data, **_kwargs)
            f.close()
        elif mode == "target":
            _kwargs = dict(date=20010101, generatingProcessIdentifier=255, param="t", level=850)
            _kwargs.update(levtype)
            to_target("file", path, metadata=_kwargs, values=data)

        ds = earthkit.data.from_source("file", path)

        assert ds[0].metadata("date") == 20010101
        assert ds[0].metadata("edition") == 2
        assert ds[0].metadata("level") == 850
        assert ds[0].metadata("levtype") == "pl"
        assert ds[0].metadata("param") == "t"
        assert ds[0].metadata("generatingProcessIdentifier") == 255
        assert ds[0].metadata("gridType") == "reduced_gg"
        assert ds[0].metadata("N") == 96
        assert ds[0].metadata("isOctahedral") == 1

        assert np.allclose(ds[0].to_numpy(), data, rtol=EPSILON, atol=EPSILON)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="ignore_cleanup_errors requires Python 3.10 or later",
)
@pytest.mark.parametrize("mode", ["ori", "compat", "target"])
@pytest.mark.parametrize("levtype", [{}, {"levtype": "pl"}])
def test_grib_output_tp(mode, levtype):
    data = np.random.random((181, 360))

    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
        path = os.path.join(tmp, "a.grib")

        if mode == "ori":
            f = earthkit.data.new_grib_output(path, date=20010101)
            # TODO: make it work for edition=2
            f.write(data, param="tp", step=48, edition=1)
            f.close()
        elif mode == "compat":
            f = new_grib_output_compat(path, metadata=dict(date=20010101, generatingProcessIdentifier=255))
            # TODO: make it work for edition=2
            f.write(values=data, param="tp", step=48, edition=1)
            f.close()
        elif mode == "target":
            to_target(
                "file",
                path,
                metadata=dict(date=20010101, generatingProcessIdentifier=255, param="tp", step=48, edition=1),
                values=data,
            )
        ds = earthkit.data.from_source("file", path)

        assert ds[0].metadata("date") == 20010101
        assert ds[0].metadata("param") == "tp"
        assert ds[0].metadata("levtype") == "sfc"
        assert ds[0].metadata("edition") == 1
        assert ds[0].metadata("step", astype=int) == 48
        assert ds[0].metadata("generatingProcessIdentifier") == 255
        assert ds[0].metadata("gridType") == "regular_ll"
        assert ds[0].metadata("Ni") == 360
        assert ds[0].metadata("Nj") == 181

        assert np.allclose(ds[0].to_numpy(), data, rtol=EPSILON, atol=EPSILON)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="ignore_cleanup_errors requires Python 3.10 or later",
)
@pytest.mark.parametrize("mode", ["ori", "compat", "target"])
@pytest.mark.parametrize("array", [True, False])
def test_grib_output_field_template(mode, array):
    data = np.random.random((7, 12))

    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
        ds = from_source("file", earthkit_examples_file("test6.grib"))
        if array:
            ds = ds.to_fieldlist()

        assert ds[0].metadata("bitsPerValue") == 4

        path = os.path.join(tmp, "a.grib")

        if mode == "ori":
            f = earthkit.data.new_grib_output(path, template=ds[0], date=20010101)
            f.write(data, param="pt", bitsPerValue=16)
            f.close()
        elif mode == "compat":
            f = new_grib_output_compat(
                path, template=ds[0], metadata=dict(date=20010101, generatingProcessIdentifier=255)
            )
            f.write(values=data, param="pt", bitsPerValue=16)
            f.close()
        elif mode == "target":
            to_target(
                "file",
                path,
                metadata=dict(date=20010101, generatingProcessIdentifier=255, param="pt", bitsPerValue=16),
                values=data,
                template=ds[0],
            )

        ds = earthkit.data.from_source("file", path)

        assert ds[0].metadata("date") == 20010101
        assert ds[0].metadata("param") == "pt"
        assert ds[0].metadata("levtype") == "pl"
        assert ds[0].metadata("edition") == 1
        assert ds[0].metadata("generatingProcessIdentifier") == 255
        assert ds[0].metadata("bitsPerValue") == 16

        assert np.allclose(ds[0].to_numpy(), data, rtol=1e-2, atol=1e-2)


@pytest.mark.parametrize("mode", ["ori", "compat", "target"])
@pytest.mark.parametrize(
    "pattern,expected_value",
    [
        ("{shortName}", {"t": 2, "u": 2, "v": 2}),
        ("{shortName}_{level}", {"t_1000": 1, "t_850": 1, "u_1000": 1, "u_850": 1, "v_1000": 1, "v_850": 1}),
        ("{date}_{time}_{step}", {"20180801_1200_0": 6}),
        ("{date}_{time}_{step:03}", {"20180801_1200_000": 6}),
    ],
)
def test_grib_output_filename_pattern(mode, pattern, expected_value):
    ds = from_source("file", earthkit_examples_file("test6.grib"))

    with temp_directory() as tmp:
        path = os.path.join(tmp, f"{pattern}.grib")

        if mode == "ori":
            f = earthkit.data.new_grib_output(path, split_output=True)
            for x in ds:
                f.write(x.values, template=x)

            f.close()
        elif mode == "compat":
            f = new_grib_output_compat(path, split_output=True)

            for x in ds:
                f.write(values=x.values, template=x)
            f.close()
        elif mode == "target":
            to_target(
                "file-pattern",
                path,
                data=ds,
            )

        for k, count in expected_value.items():
            path = os.path.join(tmp, f"{k}.grib")
            assert os.path.exists(path)
            assert len(from_source("file", path)) == count


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest

from earthkit.data import from_source
from earthkit.data.core.fieldlist import FieldList
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import earthkit_examples_file


def _check_save_to_disk(ds, len_ref, meta_ref):
    # save to disk
    tmp = temp_file()
    ds.to_target("file", tmp.path)
    assert os.path.exists(tmp.path)
    r_tmp = from_source("file", tmp.path)
    assert len(r_tmp) == len_ref
    assert r_tmp.metadata("shortName") == meta_ref
    r_tmp = None


@pytest.mark.parametrize("mode", ["oper", "multi"])
def test_grib_concat(mode):
    ds1 = from_source("file", earthkit_examples_file("test.grib"))
    ds2 = from_source("file", earthkit_examples_file("test6.grib"))

    if mode == "oper":
        ds = ds1 + ds2
    else:
        ds = from_source("multi", ds1, ds2)

    # check metadata
    assert len(ds) == 8
    md = ds1.metadata("param") + ds2.metadata("param")
    assert ds.metadata("param") == md

    # check slice
    r = ds[1]
    assert r.metadata("param") == "msl"

    r = ds[1:3]
    assert len(r) == 2
    assert r.metadata("param") == ["msl", "t"]
    assert r[0].metadata("param") == "msl"
    assert r[1].metadata("param") == "t"

    # check sel
    r = ds.sel(param="2t")
    assert len(r) == 1
    assert r.metadata("param") == ["2t"]
    assert r[0].metadata("param") == "2t"

    # check saving to disk
    _check_save_to_disk(ds, 8, md)


@pytest.mark.parametrize("mode", ["oper", "multi"])
def test_grib_concat_3a(mode):
    ds1 = from_source("file", earthkit_examples_file("test.grib"))
    ds2 = from_source("file", earthkit_examples_file("test6.grib"))
    ds3 = from_source("file", earthkit_examples_file("tuv_pl.grib"))
    md = ds1.metadata("param") + ds2.metadata("param") + ds3.metadata("param")

    if mode == "oper":
        ds = ds1 + ds2
        ds = ds + ds3
    else:
        ds = from_source("multi", ds1, ds2)
        ds = from_source("multi", ds, ds3)

    assert len(ds) == 26
    assert ds.metadata("param") == md
    _check_save_to_disk(ds, 26, md)


@pytest.mark.parametrize("mode", ["oper", "multi"])
def test_grib_concat_3b(mode):
    ds1 = from_source("file", earthkit_examples_file("test.grib"))
    ds2 = from_source("file", earthkit_examples_file("test6.grib"))
    ds3 = from_source("file", earthkit_examples_file("tuv_pl.grib"))
    md = ds1.metadata("param") + ds2.metadata("param") + ds3.metadata("param")

    if mode == "oper":
        ds = ds1 + ds2 + ds3
    else:
        ds = from_source("multi", ds1, ds2, ds3)

    assert len(ds) == 26
    assert ds.metadata("param") == md
    _check_save_to_disk(ds, 26, md)


@pytest.mark.parametrize("mode", ["oper", "multi"])
def test_grib_concat_mixed(mode):
    ds1 = from_source("file", earthkit_examples_file("test.grib"))
    ds2 = ds1.to_fieldlist()
    md = ds1.metadata("param") + ds2.metadata("param")

    if mode == "oper":
        ds = ds1 + ds2
    else:
        ds = from_source("multi", ds1, ds2)

    assert len(ds) == 4
    assert ds.metadata("param") == md
    _check_save_to_disk(ds, 4, md)


def test_grib_from_empty_1():
    ds_e = FieldList()
    ds = from_source("file", earthkit_examples_file("test.grib"))
    md = ds.metadata("param")

    ds1 = ds_e + ds
    assert id(ds1) == id(ds)
    assert len(ds1) == 2
    _check_save_to_disk(ds1, 2, md)


def test_grib_from_empty_2():
    ds_e = FieldList()
    ds = from_source("file", earthkit_examples_file("test.grib"))
    md = ds.metadata("param")

    ds1 = ds + ds_e
    assert id(ds1) == id(ds)
    assert len(ds1) == 2
    _check_save_to_disk(ds1, 2, md)


def test_grib_from_empty_3():
    ds_e = FieldList()
    ds1 = from_source("file", earthkit_examples_file("test.grib"))
    ds2 = from_source("file", earthkit_examples_file("test6.grib"))
    md = ds1.metadata("param") + ds2.metadata("param")

    ds3 = ds_e + ds1 + ds2
    assert len(ds3) == 8
    _check_save_to_disk(ds3, 8, md)


# See github issue #588
def test_grib_concat_large():
    ds_e = from_source("empty")
    ds1 = from_source("file", earthkit_examples_file("test.grib"))

    for _ in range(2000):
        ds_e += ds1.sel(param="msl")

    assert len(ds_e) == 2000


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import functools
import getpass
import logging
import os
import tempfile
import warnings
from abc import ABCMeta
from abc import abstractmethod
from contextlib import contextmanager
from typing import Callable

import deprecation
import yaml

from earthkit.data import __version__ as VERSION
from earthkit.data.utils.html import css
from earthkit.data.utils.humanize import as_bytes
from earthkit.data.utils.humanize import as_percent
from earthkit.data.utils.humanize import as_seconds
from earthkit.data.utils.humanize import interval_to_human
from earthkit.data.utils.humanize import list_to_human
from earthkit.data.utils.interval import Interval

LOG = logging.getLogger(__name__)

OLD_CONFIG_FILE = os.path.expanduser("~/.earthkit_data/settings.yaml")
CONFIG_DIR = os.path.expanduser("~/.config/earthkit/data")
CONFIG_FILE_NAME = "config.yaml"
CONFIG_FILE_ENV_NAME = "CONFIG_FILE"
ENV_PREFIX = "EARTHKIT_DATA_"


class Validator(metaclass=ABCMeta):
    @abstractmethod
    def check(self, value):
        pass

    #
    @abstractmethod
    def explain(self):
        pass

    @staticmethod
    def make(value):
        v = _validators.get(type(value), None)
        if v is not None:
            return v(value)
        else:
            raise TypeError(f"Cannot create Validator for type={type(value)}")


class ValueValidator(Validator):
    def __init__(self, value):
        self.value = value

    def check(self, value):
        return value == self.value

    def explain(self):
        return f"Valid when = {self.value}."


class IntervalValidator(Validator):
    def __init__(self, interval):
        self.interval = interval

    def check(self, value):
        return value in self.interval

    def explain(self):
        return f"Valid when {interval_to_human(self.interval)}."


class ListValidator(Validator):
    def __init__(self, values):
        self.values = values

    def check(self, value):
        return value in self.values

    def explain(self):
        return f"Valid values: {list_to_human(self.values)}."


_validators = {Interval: IntervalValidator, bool: ValueValidator, list: ListValidator}


class ConfigOption:
    def __init__(
        self,
        default,
        description,
        getter=None,
        none_ok=False,
        kind=None,
        docs_default=None,
        validator=None,
    ):
        self.default = default
        self.description = description
        self.getter = getter
        self.none_ok = none_ok
        self.kind = kind if kind is not None else type(default)
        self.docs_default = docs_default if docs_default is not None else self.default
        self.validator = validator

    def kind(self):
        return type(self.default)

    def save(self, name, value, f):
        for n in self.docs_description.split("\n"):
            print(f"# {n.strip()}", file=f)
        print(file=f)
        comment = yaml.dump({name: self.default}, default_flow_style=False)
        for n in comment.split("\n"):
            if n:
                print(f"# {n}", file=f)
        if value != self.default:
            print(file=f)
            yaml.dump({name: value}, f, default_flow_style=False)

    @property
    def docs_description(self):
        d = self.description
        t = ""
        if self.validator:
            t = self.validator.explain()

        return d.replace("{validator}", t)

    def validate(self, name, value):
        if self.validator is not None and not self.validator.check(value):
            raise ValueError(f"Config option '{name}' cannot be set to {value}. {self.validator.explain()}")


_ = ConfigOption


CONFIG_AND_HELP = {
    "user-cache-directory": _(
        os.path.join(tempfile.gettempdir(), "earthkit-data-%s" % (getpass.getuser(),)),
        """Cache directory used when ``cache-policy`` is ``user``.
        See :doc:`/guide/caching` for more information.""",
        docs_default=os.path.join("TMP", "earthkit-data-%s" % (getpass.getuser(),)),
    ),
    "temporary-cache-directory-root": _(
        None,
        """Parent of the cache directory when ``cache-policy`` is ``temporary``.
        See :doc:`/guide/caching` for more information.""",
        getter="_as_str",
        none_ok=True,
    ),
    "temporary-directory-root": _(
        None,
        """Parent of the temporary directory when ``cache-policy`` is ``off``.
        See :doc:`/guide/caching` for more information.""",
        getter="_as_str",
        none_ok=True,
    ),
    "number-of-download-threads": _(
        5,
        """Number of threads used to download data.""",
        getter="_as_int",
    ),
    "cache-policy": _(
        "off",
        """Caching policy. {validator}
        See :doc:`/guide/caching` for more information. """,
        validator=ListValidator(["off", "temporary", "user"]),
    ),
    "use-message-position-index-cache": _(
        False,
        "Stores message offset index for GRIB/BUFR files in the cache.",
    ),
    "maximum-cache-size": _(
        None,
        """Maximum disk space used by the earthkit-data cache (e.g.: 100G or 2T).
        Can be set to None.""",
        getter="_as_bytes",
        none_ok=True,
    ),
    "maximum-cache-disk-usage": _(
        "95%",
        """Disk usage threshold after which earthkit-data expires older cached
        entries (% of the full disk capacity). Can be set to None.
        See :ref:`caching` for more information.""",
        getter="_as_percent",
        none_ok=True,
    ),
    "url-download-timeout": _(
        "30s",
        """Timeout when downloading from an url.""",
        getter="_as_seconds",
    ),
    "check-out-of-date-urls": _(
        True,
        "Perform a HTTP request to check if the remote version of a cache file has changed",
    ),
    "download-out-of-date-urls": _(
        False,
        "Re-download URLs when the remote version of a cached file as been changed",
    ),
    "use-standalone-mars-client-when-available": _(
        True,
        "Use the standalone mars client when available instead of using the web API.",
    ),
    "reader-type-check-bytes": _(
        64,
        """Number of bytes read from the beginning of a source to identify its type.
        {validator}""",
        getter="_as_int",
        validator=IntervalValidator(Interval(8, 4096)),
    ),
    "grib-field-policy": _(
        "persistent",
        """GRIB field management policy for fieldlists with data on disk.  {validator}
        See :doc:`/guide/misc/grib_memory` for more information.""",
        validator=ListValidator(["persistent", "temporary"]),
    ),
    "grib-handle-policy": _(
        "cache",
        """GRIB handle management policy for fieldlists with data on disk.  {validator}
        See :doc:`/guide/misc/grib_memory` for more information.""",
        validator=ListValidator(["cache", "persistent", "temporary"]),
    ),
    "grib-handle-cache-size": _(
        1,
        """Maximum number of GRIB handles cached in memory per fieldlist with data on disk.
        Used when ``grib-handle-policy`` is ``cache``.
        See :doc:`/guide/misc/grib_memory` for more information.""",
        getter="_as_int",
        none_ok=True,
    ),
    "use-grib-metadata-cache": _(
        True,
        """Use in-memory cache kept in each field for GRIB metadata access in
        fieldlists with data on disk.
        See :doc:`/guide/misc/grib_memory` for more information.""",
    ),
    "grib-file-serialisation-policy": _(
        "path",
        """GRIB file serialisation policy for fieldlists with data on disk. {validator}""",
        validator=ListValidator(["path", "memory"]),
    ),
}


NONE = object()
DEFAULTS = {}
for k, v in CONFIG_AND_HELP.items():
    DEFAULTS[k] = v.default


@contextmanager
def new_config(s):
    """Context manager to create new config"""
    CONFIG._stack.append(s)
    CONFIG._notify()
    try:
        yield None
    finally:
        CONFIG._stack.pop()
        CONFIG._notify()


def forward(func):
    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        if self._stack:
            return func(self._stack[-1], *args, **kwargs)
        return func(self, *args, **kwargs)

    return wrapped


def save_config(path, config):
    LOG.debug("Saving config")
    from filelock import FileLock

    lock = path + ".lock"
    with FileLock(lock):
        with open(path, "w") as f:
            print("# This file is automatically generated", file=f)
            print(file=f)

            for k, v in sorted(config.items()):
                h = CONFIG_AND_HELP.get(k)
                if h:
                    print(file=f)
                    print("#", "-" * 76, file=f)
                    h.save(k, v, f)

            print(file=f)
            print("#", "-" * 76, file=f)
            print("# Version of earthkit-data", file=f)
            print(file=f)
            yaml.dump({"version": VERSION}, f, default_flow_style=False)
            print(file=f)
    try:
        os.unlink(lock)
    except OSError:
        pass


def env_var_name(name):
    return ENV_PREFIX + name.upper().replace("-", "_")


class Config:
    _auto_save_config = True
    _notify_enabled = True

    def __init__(self, config_yaml: str, defaults: dict, callbacks=[]):
        self._defaults = defaults
        self._config = dict(**defaults)
        self._callbacks = [c for c in callbacks]
        self._config_yaml = config_yaml
        self._pytest = None
        self._stack = []

    @forward
    def get(self, name: str, default=NONE):
        """[summary]

        Parameters
        ----------
            name: str
                [description]
            default: [type]
                [description]. Defaults to NONE.

        Returns
        -------
            [type]: [description]
        """
        if name not in CONFIG_AND_HELP:
            raise KeyError("No config name '%s'" % (name,))

        config_item = CONFIG_AND_HELP[name]

        getter, none_ok = (
            config_item.getter,
            config_item.none_ok,
        )
        if getter is None:
            getter = lambda name, value, none_ok: value  # noqa: E731
        else:
            getter = getattr(self, getter)

        st, value, _ = self._env(name)
        if not st:
            if default is NONE:
                value = self._config[name]
            else:
                value = self._config.get(name, default)

        return getter(name, value, none_ok)

    @forward
    def set(self, *args, **kwargs):
        """[summary]

        Parameters
        ----------
            name: str
                [description]
            value: [type]
                [description]
        """
        if len(args) == 0:
            assert len(kwargs) > 0
            for k, v in kwargs.items():
                self._set(k.replace("_", "-"), v)
            self._changed()
        else:
            if len(args) == 1 and isinstance(args[0], dict):
                for k, v in args[0].items():
                    self._set(k, v)
            else:
                self._set(*args, **kwargs)
        self._changed()

    def _set(self, name: str, *args, **kwargs):
        """[summary]

        Parameters
        ----------
            name: str
                [description]
            value: [type]
                [description]
        """
        if name not in CONFIG_AND_HELP:
            raise KeyError("No config option '%s'" % (name,))

        if self._has_env(name):
            msg = (
                f"Config option '{name}' is also set by environment variable '{env_var_name(name)}'."
                "The environment variable takes precedence and its value is returned when calling get()."
            )
            if self._auto_save_config:
                msg += " Still, the value set here will be saved to the config file."
            warnings.warn(msg)

        config_item = CONFIG_AND_HELP[name]

        klass = config_item.kind

        if klass in (bool, int, float, str):
            # TODO: Proper exceptions
            assert len(args) == 1
            assert len(kwargs) == 0
            value = args[0]
            value = klass(value)

        if klass is list:
            assert len(args) > 0
            assert len(kwargs) == 0
            value = list(args)
            if len(args) == 1 and isinstance(args[0], list):
                value = args[0]

        if klass is dict:
            assert len(args) <= 1
            if len(args) == 0:
                assert len(kwargs) > 0
                value = kwargs

            if len(args) == 1:
                assert len(kwargs) == 0
                value = args[0]

        getter, none_ok = (
            config_item.getter,
            config_item.none_ok,
        )
        if getter is not None:
            assert len(args) == 1
            assert len(kwargs) == 0
            value = args[0]
            # Check if value is properly formatted for getter
            getattr(self, getter)(name, value, none_ok)
        else:
            if not isinstance(value, klass):
                raise TypeError("Config option '%s' must be of type '%s'" % (name, klass))

        config_item.validate(name, value)
        self._config[name] = value

        LOG.debug(f"_set {name}={value} stack_size={len(self._stack)}")

    @forward
    def reset(self, name: str = None):
        """Reset config to default values.

        Parameters
        ----------
            name: str, optional
                The name of the config option to reset to default. If it
                does not have a default, it is removed. If `None` is passed, all config options are
                reset to their default values. Defaults to None.
        """
        if name is None:
            self._config = dict(**DEFAULTS)
        else:
            if name not in DEFAULTS:
                raise KeyError("No config option name '%s'" % (name,))

            self._config.pop(name, None)
            if name in DEFAULTS:
                if self._has_env(name):
                    msg = (
                        f"Config option '{name}' is also set by environment variable '{env_var_name(name)}'."
                        "The environment variable takes precedence and its value is returned when calling get()."
                    )
                    if self._auto_save_config:
                        msg += " Still, the value reset here will be saved to the config file."
                    warnings.warn(msg)
                self._config[name] = DEFAULTS[name]
        self._changed()

    @forward
    def __repr__(self):
        r = []
        env = self.env()
        for k, v in sorted(self._config.items()):
            config = CONFIG_AND_HELP.get(k, None)
            default = config.default if config else ""
            if k in env:
                r.append(f"{k}: ({env[k][0]}={env[k][1]}, {v}, {default})")
            else:
                r.append(f"{k}: ({v}, {default})")
        return "\n".join(r)

    @forward
    def _repr_html_(self):
        html = [css("table")]
        html.append("<table class='ek'>")

        env = self.env()
        columns = ["Name", "Value", "Default"]
        html.append("<tr><th>%s</th></tr>" % ("</th><th>".join(columns)))

        for k, v in sorted(self._config.items()):
            config = CONFIG_AND_HELP.get(k, None)
            default = config.default if config else ""
            if k in env:
                html.append(
                    "<tr><td>%s</td><td>%s=%r<br>(%r)</td><td>%r</td></tr>"
                    % (k, env[k][0], env[k][1], v, default)
                )
            else:
                html.append("<tr><td>%s</td><td>%r</td><td>%r</td></tr>" % (k, v, default))

        html.append("</table>")
        return "".join(html)

    @forward
    def dump(self):
        for k, v in sorted(self._config.items()):
            yield ((k, v, CONFIG_AND_HELP.get(k)))

    def _changed(self):
        if self._auto_save_config:
            self._save()
        self._notify()

    def _notify(self):
        if self._notify_enabled:
            for cb in self._callbacks:
                cb()

    def on_change(self, callback: Callable[[], None]):
        self._callbacks.append(callback)

    def _save(self):
        if self._config_yaml is None:
            return

        try:
            save_config(self._config_yaml, self._config)
        except Exception:
            LOG.error(
                "Cannot save earthkit-data config (%s)",
                self._config_yaml,
                exc_info=True,
            )

    @forward
    def save_as(self, path):
        try:
            save_config(path, self._config)
        except Exception:
            LOG.error(
                f"Cannot save earthkit-data config ({path})",
                exc_info=True,
            )

    def _as_bytes(self, name, value, none_ok):
        return as_bytes(value, name=name, none_ok=none_ok)

    def _as_percent(self, name, value, none_ok):
        return as_percent(value, name=name, none_ok=none_ok)

    def _as_seconds(self, name, value, none_ok):
        return as_seconds(value, name=name, none_ok=none_ok)

    def _as_str(self, name, value, none_ok):
        if value is None and none_ok:
            return None
        return str(value)

    def _as_int(self, name, value, none_ok):
        if value is None and none_ok:
            return None
        if isinstance(value, str):
            value = value.replace('"', "").replace("'", "").strip()
        return int(value)

    @forward
    def temporary(self, *args, config_yaml=None, **kwargs):
        tmp = Config(config_yaml, self._config)
        # until the tmp object is at the top of the stack we do not want
        # notify the observers
        if len(args) > 0 or len(kwargs) > 0:
            # tmp does not have any callbacks so it will not broadcast the changes
            tmp.set(*args, **kwargs)
        tmp._callbacks = self._callbacks
        return new_config(tmp)

    @property
    def autosave(self):
        return Config._auto_save_config

    @property
    @deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use config.autosave instead")
    def auto_save_settings(self):
        return self.autosave

    @autosave.setter
    def autosave(self, v):
        Config._auto_save_config = v

    @auto_save_settings.setter
    @deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use config.autosave instead")
    def auto_save_settings(self, v):
        self.autosave = v

    def env(self):
        r = {}
        for k in sorted(self._config.keys()):
            st, value, name = self._env(k)
            if st:
                r[k] = (name, value)
        return r

    @staticmethod
    def _has_env(name):
        return env_var_name(name) in os.environ

    @staticmethod
    def _env(name):
        name = env_var_name(name)
        if name in os.environ:
            return (True, os.environ.get(name), name)
        return (False, None, None)


# TODO: remove this function when the old config (aka settings) is not used anymore
def _migrate_old_config(config_file):
    old_config_file = os.path.expanduser(OLD_CONFIG_FILE)
    if os.path.exists(old_config_file):
        config = dict(**DEFAULTS)
        try:
            with open(old_config_file) as f:
                s = yaml.load(f, Loader=yaml.SafeLoader)
                if not isinstance(s, dict):
                    return False

                config.update(s)
                save_config(config_file, config)
                return True
        except Exception:
            LOG.error(
                f"Cannot migrate old settings to earthkit-data config file={config_file}",
                exc_info=True,
            )
    return False


def _init_config_file():
    config_file = None

    name = env_var_name(CONFIG_FILE_ENV_NAME)
    if name in os.environ:
        config_file = os.environ[name]
    if not config_file:
        config_file = os.path.expanduser(os.path.join(CONFIG_DIR, CONFIG_FILE_NAME))

    if config_file is None:
        LOG.error("Cannot construct earthkit-data config file path")
        return None

    try:
        config_dir = os.path.dirname(config_file)
        if not os.path.exists(config_dir):
            ori_mask = os.umask(0o077)
            os.makedirs(config_dir, 0o700)
            os.umask(ori_mask)
    except Exception:
        LOG.error(
            f"Cannot create earthkit-data config directory={config_dir}",
            exc_info=True,
        )

    if os.path.exists(config_dir):
        try:
            if not os.path.exists(config_file):
                if not _migrate_old_config(config_file):
                    save_config(config_file, DEFAULTS)
        except Exception:
            LOG.error(
                f"Cannot save settings to earthkit-data config file={config_file}",
                exc_info=True,
            )

    return config_file


def _init_config(config_yaml):
    save = False
    config = dict(**DEFAULTS)
    try:
        with open(config_yaml) as f:
            s = yaml.load(f, Loader=yaml.SafeLoader)
            if not isinstance(s, dict):
                s = {}

            config.update(s)

        # if s != config:
        #     save = True

        if config.get("version") < VERSION:
            save = True

    except Exception:
        LOG.error(
            "Cannot load earthkit-data config (%s), reverting to defaults",
            config_yaml,
            exc_info=True,
        )

    return config, save


save = False
config_yaml = _init_config_file()
config, save = _init_config(config_yaml)

CONFIG = Config(config_yaml, config)
if save:
    CONFIG._save()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import numpy as np
import pytest

from earthkit.data.core.temporary import temp_file
from earthkit.data.encoders import create_encoder
from earthkit.data.targets import to_target

NO_DEMO_ENCODER = True


@pytest.mark.skipif(NO_DEMO_ENCODER, reason="demo-encoder plugin not available")
@pytest.mark.plugin
def test_demo_encoder_plugin_1():
    with temp_file() as tmp:
        d = np.random.normal(0, 1, (10, 15))

        encoder = create_encoder("demo-encoder")
        assert encoder is not None

        d = encoder.encode(values=d)
        assert d is not None

        with open(tmp, "wb") as out:
            d.to_file(out)
        assert os.path.getsize(tmp) > 100


@pytest.mark.skipif(NO_DEMO_ENCODER, reason="demo-encoder plugin not available")
@pytest.mark.plugin
def test_demo_encoder_plugin_2():
    with temp_file() as tmp:
        d = np.random.normal(0, 1, (10, 15))
        to_target("file", tmp, values=d, encoder="demo-encoder")
        assert os.path.getsize(tmp) > 100


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.core.temporary import temp_file
from earthkit.data.encoders.grib import GribEncoder
from earthkit.data.targets import to_target
from earthkit.data.targets.file import FileTarget
from earthkit.data.testing import NO_RIOXARRAY
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_test_data_file


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"encoder": "grib"},
        {"encoder": GribEncoder()},
    ],
)
@pytest.mark.parametrize("direct_call", [True, False])
def test_target_file_grib_core_non_stream(kwargs, direct_call):
    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_file() as path:
        if direct_call:
            to_target("file", path, data=ds, **kwargs)
        else:
            ds.to_target("file", path, **kwargs)

        ds1 = from_source("file", path)
        assert len(ds) == len(ds1)
        assert ds1.metadata("shortName") == ["2t", "msl"]
        assert np.allclose(ds1.values[:, :4], vals_ref)


@pytest.mark.parametrize(
    "kwargs",
    [
        # {}, TODO: make it work
        {"encoder": "grib"},
        {"encoder": GribEncoder()},
    ],
)
@pytest.mark.parametrize("direct_call", [True, False])
def test_target_file_grib_core_stream(kwargs, direct_call):
    ds = from_source("file", earthkit_examples_file("test.grib"), stream=True)
    ds_ref = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds_ref.values[:, :4]

    with temp_file() as path:
        if direct_call:
            to_target("file", path, data=ds, **kwargs)
        else:
            ds.to_target("file", path, **kwargs)

        ds1 = from_source("file", path)
        assert len(ds_ref) == len(ds1)
        assert ds1.metadata("shortName") == ["2t", "msl"]
        assert np.allclose(ds1.values[:, :4], vals_ref)


def test_target_file_grib_append():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_file() as path:
        ds.to_target("file", path)
        ds.to_target("file", path, append=True)

        ds1 = from_source("file", path)
        assert len(ds1) == len(ds) * 2
        assert ds1.metadata("shortName") == ["2t", "msl"] * 2
        assert np.allclose(ds1.values[:2, :4], vals_ref)
        assert np.allclose(ds1.values[2:, :4], vals_ref)


def test_target_file_grib_non_append():
    ds = from_source("file", earthkit_examples_file("test.grib"))

    with temp_file() as path:
        ds.to_target("file", path)
        ds.to_target("file", path, append=False)

        ds1 = from_source("file", path)
        assert len(ds1) == 2


@pytest.mark.parametrize("suffix", ["grib", "grb", "grib1", "grib2", "grb1", "grb2"])
def test_target_file_grib_encoder_from_suffix(suffix):
    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_directory() as t_dir:
        path = os.path.join(t_dir, f"_r.{suffix}")
        ds.to_target("file", path)

        ds1 = from_source("file", path)
        assert len(ds) == len(ds1)
        assert ds1.metadata("shortName") == ["2t", "msl"]
        assert np.allclose(ds1.values[:, :4], vals_ref)


@pytest.mark.parametrize(
    "kwargs",
    [
        {"metadata": {"shortName": "2d", "bitsPerValue": 12}},
        {"encoder": "grib", "metadata": {"shortName": "2d", "bitsPerValue": 12}},
        {"encoder": GribEncoder(), "metadata": {"shortName": "2d", "bitsPerValue": 12}},
        {"encoder": GribEncoder(metadata={"shortName": "2d", "bitsPerValue": 12})},
    ],
)
def test_target_file_grib_set_metadata(kwargs):
    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_file() as path:
        ds.to_target("file", path, **kwargs)

        ds1 = from_source("file", path)
        assert len(ds) == len(ds1)
        assert ds1.metadata("shortName") == ["2d", "2d"]
        assert ds1.metadata("bitsPerValue") == [12, 12]
        assert np.allclose(ds1.values[:, :4], vals_ref, rtol=1e-1)


@pytest.mark.parametrize("per_field", [True, False])
def test_target_file_grib_direct_api_with_path(per_field):
    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_file() as path:
        target = FileTarget(path)
        if per_field:
            for f in ds:
                target.write(f)
        else:
            target.write(ds)
        target.close()

        # once closed, we cannot write anymore
        with pytest.raises(Exception):
            target.write(ds)

        with pytest.raises(Exception):
            target.close()

        with pytest.raises(Exception):
            target.flush()

        ds1 = from_source("file", path)
        assert len(ds) == len(ds1)
        assert ds1.metadata("shortName") == ["2t", "msl"]
        assert np.allclose(ds1.values[:, :4], vals_ref)


@pytest.mark.parametrize("per_field", [True, False])
def test_target_file_grib_direct_api_with_object(per_field):
    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_file() as path:
        with open(path, "wb") as fp:
            target = FileTarget(fp)
            if per_field:
                for f in ds:
                    target.write(f)
            else:
                target.write(ds)
            target.close()

            assert target.closed
            assert not fp.closed

            # once closed, we cannot write anymore
            with pytest.raises(ValueError):
                target.write(ds)

            with pytest.raises(ValueError):
                target.close()

            with pytest.raises(ValueError):
                target.flush()

            assert target.closed
            assert not fp.closed

        ds1 = from_source("file", path)
        assert len(ds) == len(ds1)
        assert ds1.metadata("shortName") == ["2t", "msl"]
        assert np.allclose(ds1.values[:, :4], vals_ref)


def test_target_file_grib_save_compat():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_file() as path:
        ds.save(path)

        ds1 = from_source("file", path)
        assert len(ds) == len(ds1)
        assert ds1.metadata("shortName") == ["2t", "msl"]
        assert np.allclose(ds1.values[:, :4], vals_ref)


def test_target_file_grib_write_compat():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_file() as path:
        with open(path, "wb") as f:
            ds.write(f)

        ds1 = from_source("file", path)
        assert len(ds) == len(ds1)
        assert ds1.metadata("shortName") == ["2t", "msl"]
        assert np.allclose(ds1.values[:, :4], vals_ref)


def test_target_file_bufr():
    ds = from_source("file", earthkit_examples_file("temp_10.bufr"))
    with temp_file() as path:
        ds.to_target("file", path)
        ds1 = from_source("file", path)
        assert len(ds1) == 10
        assert ds1[0].subset_count() == 1
        assert ds1[0].is_compressed() is False
        assert ds1[0].is_uncompressed() is False
        assert ds1[0].metadata("dataCategory") == 2
        assert ds1.metadata("dataCategory") == [2] * 10


def test_target_file_odb():
    ds = from_source("file", earthkit_examples_file("test.odb"))
    with temp_file() as path:
        ds.to_target("file", path)
        ds1 = from_source("file", path)
        df = ds1.to_pandas()
        assert len(df) == 717


def test_target_file_grib_to_netcdf_1():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    # vals_ref = ds.values[:, :4]

    with temp_file() as path:
        ds.to_target("file", path, encoder="netcdf")

        ds1 = from_source("file", path)
        assert len(ds1) == len(ds)
        assert ds1.metadata("param") == ["2t", "msl"]

        ds2 = ds1.to_xarray()
        assert "values" not in ds2.sizes


def test_target_file_grib_to_netcdf_2():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    # vals_ref = ds.values[:, :4]

    with temp_file() as path:
        ds.to_target("file", path, encoder="netcdf", earthkit_to_xarray_kwargs={"flatten_values": True})

        ds1 = from_source("file", path)
        ds2 = ds1.to_xarray()

        for name in ["2t", "msl"]:
            assert name in ds2.data_vars

        for name in ["latitude", "longitude"]:
            assert name in ds2.coords

        assert "values" in ds2.sizes


@pytest.mark.skipif(NO_RIOXARRAY, reason="rioxarray not available")
@pytest.mark.with_proj
def test_target_file_grib_to_geotiff():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_file() as path:
        ds.to_target("file", path, encoder="geotiff")

        ds1 = from_source("file", path)
        assert len(ds1) == len(ds)
        from earthkit.data.readers.geotiff import GeoTIFFField

        assert isinstance(ds1[0], GeoTIFFField)
        assert np.allclose(ds1.values[:, :4], vals_ref)


@pytest.mark.skipif(NO_RIOXARRAY, reason="rioxarray not available")
@pytest.mark.with_proj
def test_target_file_geotiff():
    ds = from_source("file", earthkit_test_data_file("dgm50hs_col_32_368_5616_nw.tif"))
    assert len(ds) == 3

    with temp_file() as path:
        ds.to_target("file", path)

        ds1 = from_source("file", path)
        assert len(ds) == len(ds1)
        from earthkit.data.readers.geotiff import GeoTIFFField

        assert isinstance(ds[0], GeoTIFFField)


@pytest.mark.xfail(reason="Not implemented")
@pytest.mark.skipif(NO_RIOXARRAY, reason="rioxarray not available")
@pytest.mark.with_proj
def test_target_file_geotiff_to_netcdf():
    ds = from_source("file", earthkit_test_data_file("dgm50hs_col_32_368_5616_nw.tif"))
    assert len(ds) == 3

    with temp_file() as path:
        ds.to_target("file", path)

        assert os.path.exists(path)


def test_target_file_xarray_to_netcdf():
    ds_in = from_source("file", earthkit_examples_file("test.grib"))
    # vals_ref = ds.values[:, :4]

    with temp_file() as path:
        ds = ds_in.to_xarray(flatten_values=True)

        to_target("file", path, data=ds, encoder="netcdf")

        ds1 = from_source("file", path)
        ds2 = ds1.to_xarray()

        for name in ["2t", "msl"]:
            assert name in ds2.data_vars

        for name in ["latitude", "longitude"]:
            assert name in ds2.coords

        assert "values" in ds2.sizes


def test_writers_core():
    # from earthkit.data.targets.file import FileTarget
    # from earthkit.data.targets import Target

    # from earthkit.data.targets import find_target
    # from earthkit.data.writers import to_target
    # from earthkit.data.writers import write

    # assert Target
    # assert find_target
    # assert write

    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_file() as path:
        ds.to_target("file", path)

        # assert path.exists()

        ds1 = from_source("file", path)
        assert len(ds) == len(ds1)
        assert ds1.metadata("shortName") == ["2t", "msl"]
        assert np.allclose(ds1.values[:, :4], vals_ref)

    # write("file", ds, file=temp_file())
    # write("file", ds, encoder="grib", file=temp_file())

    # target = FileTarget(temp_file())
    # target.write(ds)
    # target.write(ds, encoder="grib")

    # to_target("file", ds, file=temp_file())
    # ds.to_target("file", file=temp_file(), encoder="grib")

    # to_target("fdb", ds, conf="myconf")
    # ds.to_target("fdb", conf="myconf")
    # target = FdbTarget("myconf")
    # target.write(ds)
    # target.write(ds, format="grib")
    # target.write(ds, encoder="grib")
